__author__ = "Yuri E. Corilo"
__date__ = "Nov 11, 2019"

from threading import Thread
from pathlib import Path

from numpy import string_, array, NaN, empty
from pandas import DataFrame
import json

from corems.encapsulation.constant import Atoms
from corems.encapsulation.constant import Labels
from corems.encapsulation.output import parameter_to_dict
from corems.mass_spectrum.factory.MassSpectrumClasses import MassSpecfromFreq


class HighResMassSpecExport(Thread):
    
    '''
    TODO: add MSPeak indexes: done
    '''
    
    def __init__(self, out_file_path, mass_spectrum, output_type='excel'):
        
        '''
        output_type:str
        
            'excel', 'csv', 'hdf5' or 'pandas'
        
        '''
        Thread.__init__(self)

        self.output_file = Path(out_file_path)

        # 'excel', 'csv' or 'pandas'
        self.output_type = output_type

        self.mass_spectrum = mass_spectrum

        # collect all assigned atoms and order them accordingly to the Atoms.atoms_order list
        self.atoms_order_list = self.get_all_used_atoms_in_order(self.mass_spectrum)

        self._init_columns()

    def _init_columns(self):

        # column labels in order
        self.columns_label = ['Index',
                        'm/z',
                        'Calibrated m/z',
                        'Calculated m/z',
                        'Peak Height',
                        'Resolving Power',
                        'S/N',
                        'Ion Charge',
                        'm/z Error (ppm)',
                        'm/z Error Score',
                        'Isotopologue Similarity',
                        'Confidence Score',
                        'DBE',
                        'H/C',
                        'O/C',
                        'Heteroatom Class',
                        'Ion Type',
                        'Is Isotopologue',
                        'Mono Isotopic Index',
                        'Molecular Formula'
                        ]
                        

    @property
    def output_type(self):
        return self._output_type

    @output_type.setter
    def output_type(self, output_type):
        output_types = ['excel', 'csv', 'pandas', 'hdf5']
        if output_type in output_types:
            self._output_type = output_type
        else:
            raise TypeError(
                'Supported types are "excel", "csv" or "pandas", %s entered' % output_type)

    def save(self):
        '''wrapper to run in a separated thread'''

        if self.output_type == 'excel':
            self.to_excel()
        elif self.output_type == 'csv':
            self.to_csv()
        elif self.output_type == 'pandas':
            self.to_pandas()
        elif self.output_type == 'hdf5':
            self.to_hdf()
        else:
            raise ValueError(
                "Unkown output type: %s; it can be 'excel', 'csv' or 'pandas'" % self.output_type)

    def run(self):
        ''' run is called when the Tread start
            call exportMS.start() '''
        self.save()

    def get_pandas_df(self):

        columns = self.columns_label + self.get_all_used_atoms_in_order(self.mass_spectrum)
        dict_data_list = self.get_list_dict_data(self.mass_spectrum)
        df = DataFrame(dict_data_list, columns=columns)
        df.name = self.output_file
        return df

    def write_settings(self, output_path, mass_spectrum):

        import json

        dict_setting = parameter_to_dict.get_dict_data_ms(mass_spectrum)

        dict_setting['MassSpecAttrs'] = self.get_mass_spec_attrs(mass_spectrum)
        dict_setting['analyzer'] = mass_spectrum.analyzer
        dict_setting['instrument_label'] = mass_spectrum.instrument_label
        dict_setting['sample_name'] = mass_spectrum.sample_name

        with open(output_path.with_suffix('.json'), 'w', encoding='utf8', ) as outfile:

            output = json.dumps(dict_setting, sort_keys=True, indent=4, separators=(',', ': '))
            outfile.write(output)
    
    def to_pandas(self):
        
        columns = self.columns_label + self.get_all_used_atoms_in_order(self.mass_spectrum)

        dict_data_list = self.get_list_dict_data(self.mass_spectrum)

        df = DataFrame(dict_data_list, columns=columns)

        df.to_pickle(self.output_file.with_suffix('.pkl'))
        
        self.write_settings(self.output_file, self.mass_spectrum)
               
    def to_excel(self):

        columns = self.columns_label + self.get_all_used_atoms_in_order(self.mass_spectrum)

        dict_data_list = self.get_list_dict_data(self.mass_spectrum)

        df = DataFrame(dict_data_list, columns=columns)

        df.to_excel(self.output_file.with_suffix('.xlsx'))

        self.write_settings(self.output_file, self.mass_spectrum)

    def to_csv(self):
        
        columns = self.columns_label + self.get_all_used_atoms_in_order(self.mass_spectrum)

        dict_data_list = self.get_list_dict_data(self.mass_spectrum)

        import csv
        try:
            with open(self.output_file.with_suffix('.csv'), 'w', newline='') as csvfile:
                writer = csv.DictWriter(csvfile, fieldnames=columns)
                writer.writeheader()
                for data in dict_data_list:
                    writer.writerow(data)
            
            self.write_settings(self.output_file, self.mass_spectrum)
        
        except IOError as ioerror:
            print(ioerror)
    
    def to_json(self):

        columns = self.columns_label + self.get_all_used_atoms_in_order(self.mass_spectrum)

        dict_data_list = self.get_list_dict_data(self.mass_spectrum)

        df = DataFrame(dict_data_list, columns=columns)

        #for key, values in dict_data.items():
        #    if not values: dict_data[key] = NaN
        
        #output = json.dumps(dict_data, sort_keys=True, indent=4, separators=(',', ': '))
        return df.to_json(orient='records')

    def to_hdf(self):
        
        import h5py
        import json
        from datetime import datetime, timezone

        with h5py.File(self.output_file.with_suffix('.hdf5'), 'a') as hdf_handle:
            
            list_results = self.list_dict_to_list(self.mass_spectrum, is_hdf5= True)
            
            dict_ms_attrs = self.get_mass_spec_attrs(self.mass_spectrum)
            
            setting_dicts = parameter_to_dict.get_dict_data_ms(self.mass_spectrum)

            columns_labels = json.dumps(self.columns_label + self.get_all_used_atoms_in_order(self.mass_spectrum), sort_keys=False, indent=4, separators=(',', ': '))

            if not hdf_handle.attrs.get('date_utc'):

                timenow = str(datetime.now(timezone.utc).strftime("%d/%m/%Y %H:%M:%S %Z"))
                hdf_handle.attrs['date_utc'] = timenow
                hdf_handle.attrs['file_name'] = self.mass_spectrum.filename.name
                hdf_handle.attrs['data_structure'] = 'mass_spectrum'
                hdf_handle.attrs['analyzer'] = self.mass_spectrum.analyzer
                hdf_handle.attrs['instrument_label'] = self.mass_spectrum.instrument_label
                hdf_handle.attrs['sample_name'] = self.mass_spectrum.sample_name

            group_key = str(self.mass_spectrum.scan_number)
            
            if not group_key in hdf_handle.keys():
                
                scan_group = hdf_handle.create_group(str(self.mass_spectrum.scan_number))

                if list(self.mass_spectrum.abundance_profile):
                
                    mz_abun_array = empty(shape=(2,len(self.mass_spectrum.abundance_profile)))
                    
                    mz_abun_array[0] = self.mass_spectrum.abundance_profile
                    mz_abun_array[1] = self.mass_spectrum.mz_exp_profile
                    
                    raw_ms_dataset = scan_group.create_dataset('raw_ms', data=mz_abun_array, dtype="f8")

                else:
                    #create empy dataset for missing raw data
                    raw_ms_dataset = scan_group.create_dataset('raw_ms', dtype="f8")

                raw_ms_dataset.attrs['MassSpecAttrs'] = json.dumps(dict_ms_attrs)
                
                if isinstance(self.mass_spectrum, MassSpecfromFreq):
                    raw_ms_dataset.attrs['TransientSetting'] = json.dumps(setting_dicts.get('TransientSetting'), sort_keys=False, indent=4, separators=(',', ': '))

            else:
                
                scan_group = hdf_handle.get(group_key)
    
            # if there is not processed data len = 0, otherwise len() will return next index
            index_processed_data = str(len(scan_group.keys()))
            
            print('index_processed_data', index_processed_data)
            
            timenow = str(datetime.now(timezone.utc).strftime("%d/%m/%Y %H:%M:%S %Z"))
            
            processed_dset = scan_group.create_dataset(index_processed_data, data=list_results)
        
            processed_dset.attrs['date_utc'] = timenow

            processed_dset.attrs['ColumnsLabels'] = columns_labels
            
            processed_dset.attrs['MoleculaSearchSetting'] = json.dumps(setting_dicts.get('MoleculaSearch'), sort_keys=False, indent=4, separators=(',', ': '))
            
            processed_dset.attrs['MassSpecPeakSetting'] = json.dumps(setting_dicts.get('MassSpecPeak'), sort_keys=False, indent=4, separators=(',', ': '))

            processed_dset.attrs['MassSpectrumSetting'] = json.dumps(setting_dicts.get('MassSpectrum'), sort_keys=False, indent=4, separators=(',', ': '))

    def parameters_to_json(self):
        
        dict_setting = parameter_to_dict.get_dict_data_ms(self.mass_spectrum)

        dict_setting['MassSpecAttrs'] = self.get_mass_spec_attrs(self.mass_spectrum)
        dict_setting['analyzer'] = self.mass_spectrum.analyzer
        dict_setting['instrument_label'] = self.mass_spectrum.instrument_label
        dict_setting['sample_name'] = self.mass_spectrum.sample_name

        import re
        #pretty print 
        output = json.dumps(dict_setting)
        #output = json.dumps(dict_setting, sort_keys=False, indent=4, separators=(',', ': '))
        
        #output = re.sub(r'",\s+', '", ', output)
        
        return output

    def get_mass_spec_attrs(self, mass_spectrum):

        dict_ms_attrs = {}
        dict_ms_attrs['polarity'] = mass_spectrum.polarity
        dict_ms_attrs['rt'] =     mass_spectrum.rt
        dict_ms_attrs['tic'] =  mass_spectrum.tic
        dict_ms_attrs['mobility_scan'] =     mass_spectrum.mobility_scan
        dict_ms_attrs['mobility_rt'] =     mass_spectrum.mobility_rt
        dict_ms_attrs['Aterm'] =  mass_spectrum.Aterm
        dict_ms_attrs['Bterm'] =  mass_spectrum.Bterm
        dict_ms_attrs['Cterm'] =  mass_spectrum.Cterm
        dict_ms_attrs['baselise_noise'] =  mass_spectrum.baselise_noise
        dict_ms_attrs['baselise_noise_std'] =  mass_spectrum.baselise_noise_std

        return dict_ms_attrs


    def get_all_used_atoms_in_order(self, mass_spectrum):

        atoms_in_order = Atoms.atoms_order
        all_used_atoms = set()
        if mass_spectrum:
            for ms_peak in mass_spectrum:
                if ms_peak:
                    for m_formula in ms_peak:
                        for atom in m_formula.atoms:
                            all_used_atoms.add(atom)

        def sort_method(atom):
            return [atoms_in_order.index(atom)]

        return sorted(all_used_atoms, key=sort_method)

    def list_dict_to_list(self, mass_spectrum, is_hdf5=False):
        
        column_labels = self.columns_label + self.get_all_used_atoms_in_order(mass_spectrum)

        dict_list = self.get_list_dict_data(mass_spectrum, is_hdf5=is_hdf5)
        
        all_lines = []
        for dict_res in dict_list:
            
            result_line = [NaN] * len(column_labels)
            
            for label, value in dict_res.items():
                
                label_index = column_labels.index(label)
                result_line[label_index] =  value   
            
            all_lines.append(result_line)
        
        return  all_lines       

    def get_list_dict_data(self, mass_spectrum, include_no_match=True, include_isotopologues=True,
                           isotopologue_inline=True, no_match_inline=False, is_hdf5=False):

        dict_data_list = []

        if is_hdf5:
            encode = ".encode('utf-8')"
        else:
            encode = ""

        def add_no_match_dict_data(index, ms_peak):

            dict_result = {'Index': index,
                           'm/z':  ms_peak._mz_exp,
                           'Calibrated m/z': ms_peak.mz_exp,
                           'Peak Height': ms_peak.abundance,
                           'Resolving Power': ms_peak.resolving_power,
                           'S/N':  ms_peak.signal_to_noise,
                           'Ion Charge': ms_peak.ion_charge,
                           'Heteroatom Class' : eval("Labels.unassigned{}".format(encode)),
                           }

            dict_data_list.append(dict_result)

        def add_match_dict_data(index, ms_peak, mformula):

            formula_dict = mformula.to_dict()

            dict_result = {'Index': index,
                           'm/z':  ms_peak._mz_exp,
                           'Calibrated m/z': ms_peak.mz_exp,
                           'Calculated m/z': mformula.mz_calc,
                           'Peak Height': ms_peak.abundance,
                           'Resolving Power': ms_peak.resolving_power,
                           'S/N':  ms_peak.signal_to_noise,
                           'Ion Charge': ms_peak.ion_charge,
                           'm/z Error (ppm)': mformula.mz_error,
                           'Confidence Score': mformula.confidence_score,
                           'Isotopologue Similarity': mformula.isotopologue_similarity,
                           'm/z Error Score': mformula.average_mz_error_score,
                           'DBE': mformula.dbe,
                           'Heteroatom Class': eval("mformula.class_label{}".format(encode)),
                           'H/C': mformula.H_C,
                           'O/C': mformula.O_C,
                           'Ion Type': eval("mformula.ion_type.lower(){}".format(encode)),
                           'Is Isotopologue': int(mformula.is_isotopologue),
                           'Molecular Formula': eval("mformula.string{}".format(encode))
                           }

            if mformula.is_isotopologue:
                dict_result['Mono Isotopic Index'] = mformula.mspeak_index_mono_isotopic

            for atom in self.atoms_order_list:
                if atom in formula_dict.keys():
                    dict_result[atom] = formula_dict.get(atom)

            dict_data_list.append(dict_result)

        score_methods = mass_spectrum.molecular_search_settings.score_methods
        selected_score_method = mass_spectrum.molecular_search_settings.output_score_method

        if selected_score_method in score_methods:

            # temp set score method as the one chosen in the output
            current_method = mass_spectrum.molecular_search_settings.score_method
            mass_spectrum.molecular_search_settings.score_method = selected_score_method

            for index, ms_peak in enumerate(mass_spectrum):

                # print(ms_peak.mz_exp)

                if ms_peak:

                    m_formula = ms_peak.best_molecular_formula_candidate

                    if m_formula:

                        if not m_formula.is_isotopologue:

                            add_match_dict_data(index, ms_peak, m_formula)

                            for iso_mspeak_index, iso_mf_formula in m_formula.mspeak_mf_isotopologues_indexes:
                                iso_ms_peak = mass_spectrum[iso_mspeak_index]
                                add_match_dict_data(iso_mspeak_index, iso_ms_peak, iso_mf_formula)
                else:

                    if include_no_match and no_match_inline:
                        add_no_match_dict_data(index, ms_peak)

            if include_no_match and not no_match_inline:

                for index, ms_peak in enumerate(mass_spectrum):
                    if not ms_peak:
                        add_no_match_dict_data(index, ms_peak)     

            # reset score method as the one chosen in the output
            mass_spectrum.molecular_search_settings.score_method = current_method

        else:

            for index, ms_peak in enumerate(mass_spectrum):

                # check if there is a molecular formula candidate for the msPeak

                if ms_peak:
                    # m_formula = ms_peak.molecular_formula_lowest_error
                    for m_formula in ms_peak:

                        if mass_spectrum.molecular_search_settings.output_min_score > 0:

                            if m_formula.confidence_score >= mass_spectrum.molecular_search_settings.output_min_score:

                                if m_formula.is_isotopologue:  # isotopologues inline
                                    if include_isotopologues and isotopologue_inline:
                                        add_match_dict_data(index, ms_peak, m_formula)
                                else:
                                    add_match_dict_data(index, ms_peak, m_formula)  # add monoisotopic peak

                            # cutoff because of low score
                            else:
                                add_no_match_dict_data(index, ms_peak)

                        else:
                            if m_formula.is_isotopologue:  # isotopologues inline
                                if include_isotopologues and isotopologue_inline:
                                    add_match_dict_data(index, ms_peak, m_formula)
                            else:
                                add_match_dict_data(index, ms_peak, m_formula)  # add monoisotopic peak
                else:
                    # include not_match
                    if include_no_match and no_match_inline:
                        add_no_match_dict_data(index, ms_peak)

            if include_isotopologues and not isotopologue_inline:
                for index, ms_peak in enumerate(mass_spectrum):
                    for m_formula in ms_peak:
                        if m_formula.is_isotopologue:
                            if m_formula.confidence_score >= mass_spectrum.molecular_search_settings.output_min_score:
                                add_match_dict_data(index, ms_peak, m_formula)

            if include_no_match and not no_match_inline:
                for index, ms_peak in enumerate(mass_spectrum):
                    if not ms_peak:
                        add_no_match_dict_data(index, ms_peak)

        # remove duplicated add_match data possibly introduced on the output_score_filter step
        res = []
        [res.append(x) for x in dict_data_list if x not in res]

        return res



__author__ = 'Yuri E. Corilo'
__version__ = '1.1.10'
__doc__ = '''
[![DOI](https://zenodo.org/badge/265072913.svg)](https://zenodo.org/badge/latestdoi/265072913)

# Table of Contents  
- Introduction
  - [CoreMS](#CoreMS)  
  - [Current Version](#current-version)  
  - [Contact Information](#main-developers/contact )  
  - [Data Input](#data-input-formats)  
  - [Data Output](#data-output-formats)  
  - [Data Structure](#data-structure-types)  
  - [Features](#available-features)  
  - [Overview Slides](examples/CoreMS-Overview.pdf)
  - [Framework Documentation](https://emsl-computing.github.io/CoreMS/)
- Installation  
  - [Installation](#corems-installation)  
  - [Thermo Raw File on Mac and Linux](#thermo-raw-file-access)  
- Execution:     
  - [Jupyter Notebook and Docker containers](#molecular-database-and-jupyter-notebook-containers)  
  - [Simple Example](#simple-script-example)  
  - [Python Examples](examples/examples)
  - [Jupyter Notebook Examples](examples/notebooks)  
  

- Sibling Projects:     
    - [EnviroMS](https://github.com/EMSL-Computing/EnviroMS)  
    - [MetaMS](https://github.com/EMSL-Computing/MetaMS)  

# CoreMS

**CoreMS** is a comprehensive mass spectrometry framework for software development and data analysis of small molecules analysis.

Data handling and software development for modern mass spectrometry (MS) is an interdisciplinary endeavor requiring skills in computational science and a deep understanding of MS. To enable scientific software development to keep pace with fast improvements in MS technology, we have developed a Python software framework named CoreMS. The goal of the framework is to provide a fundamental, high-level basis for working with all mass spectrometry data types, allowing custom workflows for data signal processing, annotation, and curation. The data structures were designed with an intuitive, mass spectrometric hierarchical structure, thus allowing organized and easy access to the data and calculations. Moreover, CoreMS supports direct access for almost all vendors’ data formats, allowing for the centralization and automation of all data processing workflows from the raw signal to data annotation and curation.

- reproducible pipeline
- logical mass spectrometric data structure
- self-containing data and metadata storage
- modern molecular formulae assignment algorithms
- dynamic molecular search space database search and generator

## Current Version

### `1.1.10`

## Main Developers/Contact 
- [Yuri. E. Corilo](mailto:corilo@pnnl.gov)  
- [William Kew](mailto:william.kew@pnnl.gov)


### Data input formats

- Bruker Solarix (CompassXtract)
- Bruker Solarix transients, ser and fid (FT magnitude mode only)
- ThermoFisher (.raw)
- Spectroswiss signal booster data-acquisition station (.hdf5)
- MagLab ICR data-acquisition station (FT and magnitude mode) (.dat)
- ANDI NetCDF for GC-MS (.cdf)
- Generic mass list in profile and centroid mde (include all delimiters types and Excel formats)
- CoreMS exported processed mass list files(excel, .csv, .txt, pandas dataframe as .pkl)
- CoreMS self-containing Hierarchical Data Format (.hdf5)
- Pandas Dataframe

- Support for Could Storage using s3path.S3path  
    see examples of usage here:
      - [S3 Support](tests/s3_test.py)

### Data output formats

- Pandas data frame (can be saved using pickle, h5, etc)
- Text Files (.csv, tab separated .txt, etc)
- Microsoft Excel (xlsx)
- Automatic JSON for metadata storage and reuse
- Self-containing Hierarchical Data Format (.hdf5) including raw data and ime-series data-point for processed data-sets with all associated metadata stored as json attributes

### Data structure types

- LC-MS
- GC-MS
- IMS-MS (`TODO`)
- LC-IMS-MS (`TODO`)
- Collections (`TODO`)
- Transient
- Mass Spectra
- Mass Spectrum
- Mass Spectral Peak
- Molecular Formula
- Molecular Structure (`TODO`)

---
## Available features

### FT-MS Signal Processing

- Apodization, Zerofilling, and Magnitude mode FT
- Manual and automatic noise threshold calculation
- Peak picking using apex quadratic fitting
- Experimental resolving power calculation

### GC-MS Signal Processing

- Baseline detection, subtraction, smoothing 
- m/z based Chromatogram Peak Deconvolution,
- Manual and automatic noise threshold calculation
- First and second derivatives peak picking methods
- Peak Area Calculation

### GC-MS Calibration

- Retention Index Calibration

### GC-MS Compound Identification

- Automatic local (SQLite) or external (MongoDB or PostgreSQL) database check, generation, and search
- Automatic molecular match algorithm with all spectral similarity methods 

### FT-MS Calibration

- Frequency and m/z domain calibration functions:
- LedFord equation [ref]
- Linear equation
- Quadratic equation
- Automatic search most abundant **Ox** homologue series
- Step fit ('walking calibration") based on the LedFord equation [ref]

### FT-MS Molecular formulae search and assignment

- Automatic local (SQLite) or external (PostgreSQL) database check, generation, and search
- Automatic molecular formulae assignments algorithm for ESI(-) MS for natural organic matter analysis
- Automatic fine isotopic structure calculation and search for all isotopes
- Flexible Kendrick normalization base
- Kendrick filter using density-based clustering
- Kendrick classification
- Heteroatoms classification and visualization

### High Resolution Mass spectrum simulations

- Peak shape (Lorentz,  Gaussian, Voigt, and pseudo-Voigt)
- Peak fitting for peak shape definition
- Peak position in function of data points, signal to noise and resolving power (Lorentz and Gaussian)
- Prediction of mass error distribution
- Calculated ICR Resolving Power based on magnetic field (B), and transient time(T)

---
## CoreMS Installation 
    
```bash
pip install corems
```

By default the molecular formula database will be generated using SQLite

To use Postgresql the easiest way is to build a docker container:

```bash
docker-compose up -d
```

-  Change the url_database on MSParameters.molecular_search.url_database to:

    "postgresql+psycopg2://coremsappdb:coremsapppnnl@localhost:5432/coremsapp"

-  Set the url_database env variable COREMS_DATABASE_URL to:

    "postgresql+psycopg2://coremsappdb:coremsapppnnl@localhost:5432/coremsapp"

---
## Thermo Raw File Access:

To be able to open thermo file a installation of pythonnet is needed:
- Windows: 
    ```bash
    pip install pythonnet
    ```

- Mac and Linux:
    ```bash
    brew install mono
    pip install pythonnet   
    ```  

---
### Another option is to run the docker stack that will start the CoreMS containers:  

---

## Molecular Database and Jupyter Notebook Containers

A docker container containing:
- A custom python distribution will all dependencies installed
- A Jupyter notebook server with workflow examples
- A PostgreSQL database for the molecular formulae assignment

If you don't have docker installed, the easiest way is to [install docker for desktop](https://hub.docker.com/?overlay=onboarding)

1. Start the containers using docker-compose (easiest way): 

    On docker-compose-jupyter.yml there is a volume mapping for the tests_data directory with the data provided for testing, to change to your data location: 
    
    - locate the volumes on docker-compose-jupyter.yml:

    ```bash
    volumes:
      - ./tests/tests_data:/home/CoreMS/data
    ```
    - change "./tests/tests_data" to your data directory location

    ```bash
    volumes:
      - path_to_your_data_directory:/home/corems/data
    ```
    - save the file and then call:
    
    ```bash
    docker-compose -f docker-compose-jupyter.yml up
    ```

2. Another option is to manually build the containers: 

    - Build the corems image:
        ```bash
        docker build -t corems:local .
        ```
    - Start the database container:
        ```bash
        docker-compose up -d   
        ```
    - Start the Jupyter Notebook:
        ```bash
        docker run --rm -v ./data:/home/CoreMS/data corems:local
        ```
    
    - Open your browser, copy and past the URL address provided in the terminal: `http://localhost:8888/?token=<token>.`

    - Open the CoreMS-Tutorial.ipynb

___
## Simple Script Example

More examples can be found under the directory docs/example, docs/notebooks

- Basic functionality example

```python
from corems.transient.input.brukerSolarix import ReadBrukerSolarix
from corems.molecular_id.search.molecularFormulaSearch import SearchMolecularFormulas
from corems.mass_spectrum.output.export import HighResMassSpecExport
from matplotlib import pyplot

file_path= 'tests/tests_data/ESI_NEG_SRFA.d'

#Bruker Solarix class reader
bruker_reader = ReadBrukerSolarix(file_path)

#access the transient object
bruker_transient_obj = bruker_reader.get_transient()

#calculates the transient duration time
T =  bruker_transient_obj.transient_time

#access the mass spectrum object
mass_spectrum_obj = bruker_transient_obj.get_mass_spectrum(plot_result=False, auto_process=True)

# - search monoisotopic molecular formulas for all mass spectral peaks
# - calculate fine isotopic structure based on monoisotopic molecular formulas found and current dynamic range
# - search molecular formulas of correspondent calculated isotopologues,
# - settings are stored at SearchConfig.json and can be changed directly on the file or inside the framework class

SearchMolecularFormulas(mass_spectrum_obj, first_hit=False).run_worker_mass_spectrum()

# iterate over mass spectral peaks objs
for mspeak in mass_spectrum_obj.sort_by_abundance():

    # returns true if there is at least one molecular formula associated
    # with the mass spectral peak
    # same as mspeak.is_assigned -- > bool
    if  mspeak:

        # get the molecular formula with the highest mass accuracy
        molecular_formula = mspeak.molecular_formula_lowest_error

        # plot mz and peak height, use mass_spectrum_obj.mz_exp to access all mz
        # and mass_spectrum_obj.mz_exp_profile to access mz with all available datapoints
        pyplot.plot(mspeak.mz_exp, mspeak.abundance, 'o', c='g')

        # iterate over all molecular formulae associated with the ms peaks obj
        for molecular_formula in mspeak:

            #check if the molecular formula is a isotopologue
            if molecular_formula.is_isotopologue:

                #access the molecular formula text representation
                print (molecular_formula.string)

                #get 13C atoms count
                print (molecular_formula['13C'])
    else:
        #get mz and peak height
        print(mspeak.mz_exp,mspeak.abundance)


#exporting data
mass_spectrum_obj.to_csv("filename")

mass_spectrum_obj.to_hdf("filename")
# save pandas Datarame to pickle
mass_spectrum_obj.to_pandas("filename")
# get pandas Dataframe
df = mass_spectrum_obj.to_dataframe()
```
## Citing CoreMS

If you use CoreMS in your work, please use the following citation:
Version [1.1.10 Release on GitHub](https://github.com/EMSL-Computing/CoreMS/releases/tag/1.1.10), archived on Zenodo:  

[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.4641553.svg)](https://doi.org/10.5281/zenodo.4641553)
```
Yuri E. Corilo, William R. Kew, Lee Ann McCue. (2021, March 27). EMSL-Computing/CoreMS: CoreMS 1.1.10 (Version v1.1.10), as developed on Github. Zenodo. http://doi.org/10.5281/zenodo.4641553
```
## Disclaimer

This material was prepared as an account of work sponsored by an agency of the
United States Government.  Neither the United States Government nor the United
States Department of Energy, nor Battelle, nor any of their employees, nor any
jurisdiction or organization that has cooperated in the development of these
materials, makes any warranty, express or implied, or assumes any legal
liability or responsibility for the accuracy, completeness, or usefulness or
any information, apparatus, product, software, or process disclosed, or
represents that its use would not infringe privately owned rights.

Reference herein to any specific commercial product, process, or service by
trade name, trademark, manufacturer, or otherwise does not necessarily
constitute or imply its endorsement, recommendation, or favoring by the United
States Government or any agency thereof, or Battelle Memorial Institute. The
views and opinions of authors expressed herein do not necessarily state or
reflect those of the United States Government or any agency thereof.

                 PACIFIC NORTHWEST NATIONAL LABORATORY
                              operated by
                                BATTELLE
                                for the
                   UNITED STATES DEPARTMENT OF ENERGY
                    under Contract DE-AC05-76RL01830
'''
import time
import os
import sys
import hashlib

		
def timeit(method):
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        if 'log_time' in kw:
            name = kw.get('log_name', method.__name__.upper())
            kw['log_time'][name] = int((te - ts) * 1000)
        else:
            print( "%r  %2.2f ms" % (method.__name__, (te - ts) * 1000))
        return result
    return timed


class SuppressPrints:
    
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout  

def get_filename(app=None):
    
    from PySide2.QtCore import Qt, QCoreApplication
    from PySide2.QtWidgets import QApplication, QFileDialog 
    from pathlib import Path

    app = QApplication(sys.argv)
    file_dialog = QFileDialog()
    file_dialog.setWindowFlags(Qt.WindowStaysOnTopHint)
    file_location, _ = file_dialog.getOpenFileName()
    
    if file_location:
        QCoreApplication.processEvents()
        return Path(file_location)
    
    else:
        
        return None

def get_dirname(app=None):
    
    from PySide2.QtCore import Qt, QCoreApplication
    from PySide2.QtWidgets import QApplication, QFileDialog 
    from pathlib import Path

    app = QApplication(sys.argv)
    file_dialog = QFileDialog()
    file_dialog.setWindowFlags(Qt.WindowStaysOnTopHint)
    file_location = file_dialog.getExistingDirectory()
    
    if file_location:
        QCoreApplication.processEvents()
        return Path(file_location)
    
    else:
        
        return None

def get_dirnames(app=None):   
    
    from PySide2.QtCore import Qt, QCoreApplication
    from PySide2.QtWidgets import QApplication, QFileDialog, QTreeView, QListView, QAbstractItemView
    from pathlib import Path
    
    if not app:
        app = QApplication(sys.argv)
    #file_dialog = QFileDialog()
    #file_dialog.setWindowFlags(Qt.WindowStaysOnTopHint)
    #file_location = file_dialog.getOpenFileNames()
    
    file_dialog = QFileDialog()
    file_dialog.setFileMode(QFileDialog.DirectoryOnly)
    file_dialog.setOption(QFileDialog.DontUseNativeDialog, True)
    file_view = file_dialog.findChild(QListView, 'listView')

    # to make it possible to select multiple directories:
    if file_view:
        file_view.setSelectionMode(QAbstractItemView.MultiSelection)
    f_tree_view = file_dialog.findChild(QTreeView)
    if f_tree_view:
        f_tree_view.setSelectionMode(QAbstractItemView.MultiSelection)

    if file_dialog.exec():
        paths = file_dialog.selectedFiles()
        QCoreApplication.processEvents()
        for path in paths:
            yield Path(path)

def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

def corems_md5(fname):

    bytes_io = fname.open('rb').read()

    md5_returned = hashlib.sha256(bytes_io).hexdigest()
    
    return "{}:{}".format("sha256", md5_returned)
        



__author__ = "Yuri E. Corilo"
__date__ = "Oct 23, 2019"

from corems.mass_spectrum.factory.MassSpectrumClasses import MassSpecProfile, MassSpecCentroid
from corems.encapsulation.factory.parameters import default_parameters
from corems.encapsulation.constant import Labels

def ms_from_array_profile(mz, abundance,  dataname, polarity=-1,  auto_process=True, auto_noise=True, data_type=Labels.simulated_profile):

    data_dict = {Labels.mz: mz, Labels.abundance: abundance}
    
    output_parameters = get_output_parameters(polarity, dataname)
    
    output_parameters[Labels.label] = data_type

    ms = MassSpecProfile(data_dict, output_parameters, auto_process=auto_process, auto_noise=auto_noise)
    
    return ms

def ms_from_array_centroid(mz, abundance, rp, s2n, dataname, polarity=-1, auto_process=True):

    data_dict = {Labels.mz: mz, Labels.abundance: abundance, Labels.s2n : s2n, Labels.rp: rp}
    
    output_parameters = get_output_parameters(polarity, dataname)
        
    return MassSpecCentroid(data_dict, output_parameters)
    
def get_output_parameters(polarity, file_location):
        
        d_params = default_parameters(file_location)
        
        d_params['analyzer'] = 'Generic Simulated'

        d_params['instrument_label'] = 'Generic Simulated'

        d_params["polarity"] = polarity
        
        d_params["filename_path"] = file_location
        
        d_params["mobility_scan"] = 0
        
        d_params["mobility_rt"] = 0
        
        d_params["scan_number"] = 0
        
        d_params["rt"] = 0

        d_params[Labels.label] = Labels.simulated_profile
        
        return d_params



from corems.encapsulation.output import parameter_to_dict 


def dump_all_settings_json(filename='SettingsCoreMS.json', file_path=None):
    
    from pathlib import Path
    import json
    '''Write JSON file into current directory
    '''        
    data_dict_all = parameter_to_dict.get_dict_all_default_data()
    
    if not file_path:
        file_path = Path.cwd() / filename 
    
    with open(file_path, 'w', encoding='utf8', ) as outfile:
            
        import re
        #pretty print 
        output = json.dumps(data_dict_all, sort_keys=False, indent=4, separators=(',', ': '))
        output = re.sub(r'",\s+', '", ', output)
        
        outfile.write(output)

def dump_ms_settings_json(filename='SettingsCoreMS.json', file_path=None):
    
    from pathlib import Path
    import json
    '''Write JSON file into current directory
    '''        
    data_dict = parameter_to_dict.get_dict_ms_default_data()

    if not file_path:
        
        file_path = Path.cwd() / filename 
    
    with open(file_path, 'w', encoding='utf8', ) as outfile:
            
        import re
        #pretty print 
        output = json.dumps(data_dict, sort_keys=False, indent=4, separators=(',', ': '))
        output = re.sub(r'",\s+', '", ', output)
        
        outfile.write(output)

def dump_gcms_settings_json(filename='SettingsCoreMS.json', file_path=None):
    '''Write JSON file into current directory
    '''        
    from pathlib import Path
    import json
    
    data_dict = parameter_to_dict.get_dict_gcms_default_data()

    if not file_path:
        
        file_path = Path.cwd() / filename 
    
    with open(file_path, 'w', encoding='utf8', ) as outfile:
            
        import re
        #pretty print 
        output = json.dumps(data_dict, sort_keys=False, indent=4, separators=(',', ': '))
        output = re.sub(r'",\s+', '", ', output)
        
        outfile.write(output)       


from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
import statistics as st
import csv

from corems.mass_spectra.calc.GC_Calc import GC_Calculations
from corems.mass_spectrum.factory.MassSpectrumClasses import MassSpecBase
from corems.mass_spectrum.factory.MassSpectrumClasses import MassSpecCentroidLowRes
from corems.chroma_peak.factory.ChromaPeakClasses import GCPeak
from corems.mass_spectra.calc import SignalProcessing as sp

class LossFinderTargeted(GC_Calculations):

    def __init__(self, ref_file = 'Unknown', noise_cutoff = 'Unknown', tolerance = "Unknown"):

        self.tolerance = float()

        self.noise_cutoff = float()

        self.offset_hits = {}
        
        self.mz_filtered = {}

        self.abund_filtered = {}
        
        self.mz_count = float()
        
        #self._mz_exp = MassSpecBase._mspeaks
        #self._abundance = MassSpecBase._abundance

    def ms_info_get(self, mass_spectra_obj):

        mz_dict = {}
        abund = {}

        self.mz_count = sum([len(mz) for mz in mass_spectra_obj.values()])

        for scan_number, ms_obj in mass_spectra_obj.items():
            
            mz_dict.update({scan_number:ms_obj.mz_exp})
            abund.update({scan_number:ms_obj.abundance})
        
        return mz_dict, abund

    def loss_ref_get(self, file_location, tolerance):

        offset_ref = {}
        range_ref = {}

        with open(file_location) as ref:

            ref_reader = csv.reader(ref, delimiter=',')
            next(ref_reader)

            for name, mass_offset in ref_reader:
                offset_ref.setdefault(name,float(mass_offset))

        for key, val in offset_ref.items():

            range_ref.update({key:(val-tolerance, val+tolerance)})

        return range_ref
    
    def threshold_filter(self, mz_dict, Intensity, noise_cutoff):
        
        for scan_number, info in Intensity.items():

            cutoff = st.mean(Intensity[scan_number])*noise_cutoff

            noise = set([peak for peak in Intensity[scan_number] if peak <=cutoff])

            mz_dict[scan_number] = [mz for mz, peak in zip(mz_dict[scan_number], Intensity[scan_number]) if peak not in noise ]

            Intensity[scan_number] = [peak for peak in Intensity[scan_number] if peak >= cutoff]

        return mz_dict , Intensity
    
    def mz_pair_checker(self, chem, lower, upper, mz1, spectrum, Intensity, scan_number):

        for mz2 in spectrum:

            if mz1 > mz2 and lower <= abs(mz1-mz2) <= upper:

                if chem not in self.offset_hits.keys():
                    self.offset_hits.update( {chem:[((mz2, mz1, Intensity[spectrum.index(mz2)] , Intensity[spectrum.index(mz1)], chem, scan_number ))]} )

                else:
                    self.offset_hits[chem].append( ((mz2,mz1, Intensity[spectrum.index(mz2)] , Intensity[spectrum.index(mz1)], chem, scan_number )) )

        

    def findpeakoffset(self, range_ref, mz_filtered, abund_filtered):

        count = 0    

        for chem, ref_pair in range_ref.items():
        
            for scan_number in mz_filtered.keys():

                while count < len(mz_filtered[scan_number])-1:
                #while count < 2:    

                    self.mz_pair_checker(chem, ref_pair[0], ref_pair[1], mz_filtered[scan_number][count], mz_filtered[scan_number], abund_filtered[scan_number], scan_number)
            
                    count+=1

                #print(self.offset_hits)

                count=0

        return self.offset_hits

    def LF_out(self, LF_dict, mz_count):
        data = pd.DataFrame()

        for chem, stats in LF_dict.items():
        
            data = data.append(pd.DataFrame(stats), ignore_index=True)

        data.columns = ['mz1','mz2','Intensity1','Intensity2','Name','scan_number']

        data.to_csv('Loss_finder.csv', encoding='utf-8', index=False, header = True, mode='w')

        chem = data['Name'].value_counts().keys().tolist()
        freq = data['Name'].value_counts().tolist()

        chem.reverse()
        freq.reverse()

        plt.barh(chem,freq)
        
       
        plt.title("Chemical Loss Frequency")
        plt.xlabel("Number of Loss found")
        
        plt.tight_layout()

        ax = plt.subplot()
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)

        for n, i in enumerate(chem):
        #Create an axis text object
            plt.text(freq[n] + 12, #X location of text (with adjustment)
                n, #Y location
                s=f'{round(((freq[n])/mz_count)*100,3)}%', #Required label with formatting
                #s=f'{range_ref[n]}', #Required label with formatting
                va='center', #Vertical alignment
                color='#661D98', #Font colour and size
                fontsize=12)

        # plt.show()

    def plot_offset(self):
        #MassSpecBase.plot_mz_domain_profile(MassSpecBase)
        #MassSpecBase.plot_profile_and_noise_threshold(MassSpecBase)

        #plt.show()

        out = MassSpecBase.get_mz_and_abundance_peaks_tuples(MassSpecBase)

        return out


import time

from numpy import where, average, std, isnan, inf, hstack, median, argmax, percentile
from corems import chunks
import warnings

#from matplotlib import pyplot
__author__ = "Yuri E. Corilo"
__date__ = "Jun 27, 2019"

class NoiseThresholdCalc:

    def get_noise_threshold(self) -> ( (float, float), (float,float) ):
        ''' return two tuples (min_mz, max_mz) , (noise_threshold, noise_threshold)'''
        
        if self.is_centroid:

            x = min(self.mz_exp), max((self.mz_exp))
            
            if self.settings.threshold_method == 'auto':
                
                abundance_threshold = self.baselise_noise + (self.settings.noise_threshold_std * self.baselise_noise_std)
                y = (abundance_threshold, abundance_threshold)

            elif self.settings.threshold_method == 'signal_noise':

                normalized_threshold = (self.max_abundance * self.settings.s2n_threshold )/self.max_signal_to_noise
                y = (normalized_threshold, normalized_threshold)
            
            elif self.settings.threshold_method == "relative_abundance":

                normalized_threshold = (max(self.abundance)/100)*self.settings.relative_abundance_threshold
                y = (normalized_threshold, normalized_threshold)    
            
            else:
                    raise  Exception("%s method was not implemented, please refer to corems.mass_spectrum.calc.NoiseCalc Class" % self.settings.threshold_method)
                
            return x, y    

        else:

            if self.baselise_noise and self.baselise_noise_std:
                
                x = (self.mz_exp_profile.min(), self.mz_exp_profile.max())
                y = (self.baselise_noise_std, self.baselise_noise_std)
                
                if self.settings.threshold_method == 'auto':
                
                    #print(self.settings.noise_threshold_std)
                    abundance_threshold = self.baselise_noise + (self.settings.noise_threshold_std * self.baselise_noise_std)
                    
                    y = (abundance_threshold, abundance_threshold)

                elif self.settings.threshold_method == 'signal_noise':

                    max_sn = self.abundance_profile.max()/self.baselise_noise_std

                    normalized_threshold = (self.abundance_profile.max() * self.settings.s2n_threshold )/max_sn
                    y = (normalized_threshold, normalized_threshold)

                elif self.settings.threshold_method == "relative_abundance":

                    normalized_threshold = (self.abundance_profile.max()/100)*self.settings.relative_abundance_threshold
                    y = (normalized_threshold, normalized_threshold)
                    

                else:
                    raise  Exception("%s method was not implemented, \
                        please refer to corems.mass_spectrum.calc.NoiseCalc Class" % self.settings.threshold_method)
                
                return x, y
            
            else:
                
                warnings.warn(
                    "Noise Baseline and Noise std not specified,\
                    defaulting to 0,0 run process_mass_spec() ?"
                )    
                return (0,0) , (0,0)

    def cut_mz_domain_noise(self, auto):
        
        min_mz_whole_ms = self.mz_exp_profile.min()
        max_mz_whole_ms = self.mz_exp_profile.max()

        if auto:
            
            # this calculation is taking too long (about 2 seconds)
            number_average_molecular_weight = self.weight_average_molecular_weight(
                profile=True)
           
            # +-200 is a guess for testing only, it needs adjustment for each type of analysis
            # need to check min mz here or it will break
            min_mz_noise = number_average_molecular_weight - 100
            # need to check max mz here or it will break
            max_mz_noise = number_average_molecular_weight + 100

        else:

            min_mz_noise = self.settings.min_noise_mz
            max_mz_noise = self.settings.max_noise_mz

        if min_mz_noise < min_mz_whole_ms:
            min_mz_noise = min_mz_whole_ms

        if max_mz_noise > max_mz_whole_ms:
            max_mz_noise = max_mz_whole_ms

        #print(min_mz_noise, max_mz_noise)
        low_mz_index = (where(self.mz_exp_profile >= min_mz_noise)[0][0])
        #print(self.mz_exp_profile[low_mz_index])
        # low_mz_index = (argmax(self.mz_exp_profile <= min_mz_noise))
        
        high_mz_index = (where(self.mz_exp_profile <= max_mz_noise)[-1][-1])
        
        #high_mz_index = (argmax(self.mz_exp_profile <= max_mz_noise))
        
        if high_mz_index > low_mz_index:
            # pyplot.plot(self.mz_exp_profile[low_mz_index:high_mz_index], self.abundance_profile[low_mz_index:high_mz_index])
            # pyplot.show()
            return self.mz_exp_profile[high_mz_index:low_mz_index], self.abundance_profile[low_mz_index:high_mz_index]
        else:
            # pyplot.plot(self.mz_exp_profile[high_mz_index:low_mz_index], self.abundance_profile[high_mz_index:low_mz_index])
            # pyplot.show()
            return self.mz_exp_profile[high_mz_index:low_mz_index], self.abundance_profile[high_mz_index:low_mz_index]


    def from_posterior(self, param, samples):

        import pymc3 as pm
        import numpy as np
        import theano.tensor as tt
        from theano import as_op
        from scipy.stats import gaussian_kde
        
        smin, smax = np.min(samples), np.max(samples)
        width = smax - smin
        x = np.linspace(smin, smax, 100)
        y = gaussian_kde(samples)(x)
        
        # what was never sampled should have a small probability but not 0,
        # so we'll extend the domain and use linear approximation of density on it
        x = np.concatenate([[x[0] - 3 * width], x, [x[-1] + 3 * width]])
        y = np.concatenate([[0], y, [0]])
        
        return pm.distributions.Interpolated(param, x, y)

    def error_model_from_trace(self, trace, ymincentroid):
         
        import pymc3 as pm
        #from pymc3 import traceplot, plot_posterior
        
        with pm.Model() as model2:
            
            sd = self.from_posterior('sd', trace['sd'])
            y = pm.HalfNormal('y', sd=sd, observed=ymincentroid)
            start = pm.find_MAP()
            step = pm.NUTS() # Hamiltonian MCMC with No U-Turn Sampler
            trace = pm.sample(1000, step, start, random_seed=123, progressbar=True, tune=1000)
            pm.summary(trace)
            #plot_posterior(trace)
            #traceplot(trace)    
            return pm.summary(trace)['mean'].values[0] 

    def simple_model_error_dist(self,  ymincentroid):
        
        import pymc3 as pm
        # from pymc3 import traceplot, plot_posterior
        #import seaborn as sns
        #f, ax = pyplot.subplots(figsize=(6, 6))
        #sns.distplot(ymincentroid)
        #sns.kdeplot(ymincentroid, ax=ax, shade=True, color="g")
        #sns.rugplot(ymincentroid, color="black", ax=ax)
        #ax.set(xlabel= "Peak Minima Magnitude", ylabel= "Density")
        #pyplot.show()

        with pm.Model() as model:
            
            #mu = pm.Uniform('mu', lower=-1, upper=1)
            lower = ymincentroid.min()
            upper = ymincentroid.max()
            
            sd = pm.Uniform('sd', lower=lower , upper=upper)
            
            y = pm.HalfNormal('y', sd=sd, observed=ymincentroid)
            
            start = pm.find_MAP()
            step = pm.NUTS() # Hamiltonian MCMC with No U-Turn Sampler
            trace = pm.sample(1000, step, start, random_seed=123, progressbar=True, tune=1000)
            
            return pm.summary(trace)['mean'].values[0] 
            

    def get_noise_average(self, ymincentroid, auto=True, bayes=False):
        # assumes noise to be gaussian and estimate noise level by 
        # calculating the valley. If bayes is enable it will 
        # model the valley distributuion as half-Normal and estimate the std
        
        average_noise = median((ymincentroid))*2 if auto else median(ymincentroid)
        
        if bayes:
            
            s_deviation = self.simple_model_error_dist(ymincentroid)
        
        else:
            
            s_deviation = ymincentroid.std()*3 if auto else ymincentroid.std()
            
        return average_noise, s_deviation

    def get_abundance_minima_centroid(self, mz_cut, abun_cut):

        maximum = self.abundance_profile.max()
        threshold_min = (maximum * 1.00)

        y = -abun_cut

        dy = y[1:] - y[:-1]
        '''replaces NaN for Infinity'''
        indices_nan = where(isnan(y))[0]
        
        if indices_nan.size:

            y[indices_nan] = inf
            dy[where(isnan(dy))[0]] = inf

        
        indices = where((hstack((dy, 0)) < 0) & (hstack((0, dy)) > 0))[0]
        
        if indices.size and threshold_min is not None:
            indices = indices[abun_cut[indices] <= threshold_min]

        # pyplot.plot(mz_cut[indices], abun_cut[indices], linewidth=0, marker='o', color='red', markersize=2)
        # pyplot.show()

        # pyplot.hist(x=abun_cut[indices], bins='auto', color='#0504aa',
                            # alpha=0.7, rwidth=0.85)
        
        # cutoff = percentile(abun_cut[indices], 0.98)
        
        # print(cutoff)
        
        # pyplot.show()    
        return abun_cut[indices]


    def run_noise_threshold_calc(self, auto, bayes=False):
        
        if self.is_centroid:
            # calculates noise_baseline and noise_std
            # needed to run auto noise threshold mode
            # it is not used for signal to noise nor 
            # relative abudance methods
            abundances_chunks = chunks(self.abundance, 50)
            each_min_abund = [min(x) for x in abundances_chunks]

            return average(each_min_abund), std(each_min_abund)
        
        else:

            mz_cut, abundance_cut = self.cut_mz_domain_noise(auto)
            
            if auto:

                yminima = self.get_abundance_minima_centroid(mz_cut, abundance_cut)
                
                return self.get_noise_average(yminima, auto=auto, bayes=bayes)

            else:
                
                # pyplot.show()
                return self.get_noise_average(abundance_cut,auto=auto, bayes=bayes)





import numpy as np

from pandas import Series, DataFrame

from scipy.signal import savgol_filter, boxcar
from scipy import interpolate
from matplotlib import pyplot as plt
from numpy import abs
from numpy import array, poly1d, polyfit

def peak_detector(tic, max_tic):

    dy = derivate(tic)

    indexes = np.where((np.hstack((dy, 0)) < 0) & (np.hstack((0, dy)) > 0))[0]
    
    for index in indexes:
        
        start_index = find_minima(index, tic, right=False)
        final_index = find_minima(index, tic)

        yield (start_index, index, final_index)

def find_nearest_scan(data, nodes):

    array_data = array(nodes)

    scan_index = (abs(array_data - data)).argmin()

    return nodes[scan_index]

def peak_picking_first_derivative(domain, signal, max_height, max_prominence, max_signal, min_peak_datapoints,
                           signal_threshold=0.2, correct_baseline=True, plot_res=False):
    
    if correct_baseline:
        signal = signal - baseline_detector(signal, domain, max_height, max_prominence)

    domain = np.array(domain)
    signal = np.array(signal)

    # dy_signal = derivate(signal)

    # dydy = derivate(dy_signal)
    
    # where returns a tuple of indexes and data type and we only need  the indexes
    # right_indexes = np.where((np.hstack( (0, dy_signal)) < neg_dy_threshold))[0]
    # left_indexes = np.where((np.hstack( (0, dy_signal)) > pos_dy_threshold))[0]
  
    # dy_left_signal = derivate(left_indexes)  
    # dy_left_signal_zero_filled = np.hstack( (dy_left_signal,0))

    # start_peak = []

    # needs a more efficient way of doing that 
    # for index in range(1,len(dy_left_signal_zero_filled)-1):
    #    delta = dy_left_signal_zero_filled[index]
    #    delta_previous = dy_left_signal_zero_filled[index - 1]
    #    delta_next = dy_left_signal_zero_filled[index + 1]
    #    if delta == 1 and delta_previous != 1:
    #        start_peak.append(left_indexes[index]-1)
    # start_peak = array(start_peak)

    dy = derivate(signal)
    apex_indexes = np.where((np.hstack((dy, 0)) < 0) & (np.hstack((0, dy)) > 0))[0]
    # min_indexes = np.where((np.hstack((dy, 0)) > 0) & (np.hstack((0, dy)) < 0))[0]

    start_peak = []
    end_peak = []

    pos_dy_threshold = max(dy) * 0.0005
    neg_dy_threshold = min(dy) * 0.0005
    len_dy = len(dy)
    # take apex_index and move left to find start
    for index in apex_indexes:
        # catch for starting position
        if index == 0:
            index_start = index
        else:
            index_start = index - 1

        # catch for ending position
        if index == (len(dy) - 1):
            index_end = index
        else:
            index_end = index + 1

        # while dy[index_start-1] > 0 and index_start != 0:
        while dy[index_start - 1] > pos_dy_threshold and index_start > 0:
            index_start = index_start - 1
        start_peak.append(index_start)

        # while dy[index_end] < 0 and index_end != (len(dy) - 1):
        while dy[index_end] < neg_dy_threshold and index_end != (len_dy - 1):
            index_end = index_end + 1
        end_peak.append(index_end)

    start_peak = array(start_peak)
    end_peak = array(end_peak)

    # left_index = []
    # right_index = []

    last_closest_right = 0

    for apex_index in apex_indexes:

        index_gt_apex = np.where(end_peak >= apex_index)[0]
        index_lt_apex = np.where(start_peak <= apex_index)[0]
        # index_gt_apex = np.where(min_indexes >= apex_index)[0]
        # index_lt_apex = np.where(min_indexes <= apex_index)[0]

        if not index_gt_apex.size == 0 and not index_lt_apex.size == 0:

            closest_right = find_nearest_scan(apex_index, end_peak[index_gt_apex])
            closest_left = find_nearest_scan(apex_index,  start_peak[index_lt_apex])
            # closest_right = find_nearest_scan(apex_index, min_indexes[index_gt_apex])
            # closest_left = find_nearest_scan(apex_index, min_indexes[index_lt_apex])

            x = [closest_left, closest_right]
            y = [signal[closest_left], signal[closest_right]]

            pol = poly1d(polyfit(x, y, 1))

            corrected_peak_height = signal[apex_index] - pol(apex_index)

            if (corrected_peak_height / max_signal) * 100 > signal_threshold:

                if (closest_right - closest_left) >= min_peak_datapoints:

                    if plot_res:

                        # plt.plot(domain[closest_left: closest_right+1], dydy[closest_left:closest_right+1], c='black')
                        plt.plot(domain[closest_left: closest_right+1], dy[closest_left:closest_right+1], c='red')
                        plt.plot(domain[[apex_index, apex_index]], [signal[apex_index], pol(apex_index)], c='red')
                        plt.plot(domain[start_peak], signal[start_peak], c='blue', linewidth='0', marker="^")
                        # plt.plot(domain[[closest_left,apex_index,closest_right]], signal[[closest_left,apex_index,closest_right]], c='blue', linewidth='0', marker="s")    
                        # plt.plot(domain[[closest_left,apex_index,closest_right]], pol([closest_left, apex_index, closest_right]), c='red')
                        plt.plot(domain[closest_left: closest_right+1], signal[closest_left:closest_right+1], c='black')
                        plt.title(str((corrected_peak_height/max_signal)*100))
                        
                        # plt.show()

                    # if not closest_right <= apex_index:
                    #right_index.append(closest_right)
                    #print (closest_right, apex_index)

                    #left_index.append(closest_left)
                    #print (closest_left, apex_index)
                    #plt.show()
                    #if closest_left < last_closest_right:
                    #    closest_left = last_closest_right
                        
                    yield (closest_left, apex_index, closest_right)
                    last_closest_right = closest_right
    
    #plt.plot(domain, dydy, c='black')
    #plt.plot(domain, np.hstack((0, dy)), c='green')
    #plt.plot(domain, [0 for i in range(len(domain))], c='blue')
    #plt.plot(domain[right_index], signal[right_index], c='blue', linewidth='0', marker="s")
    #plt.plot(domain[left_index], signal[left_index], c='yellow', linewidth='0', marker="^")
    #plt.plot(domain[apex_indexes], signal[apex_indexes], c='red', linewidth='0', marker="^")
    #plt.plot(domain[right_indexes], signal[right_indexes], c='green', linewidth='0', marker="^")
    #plt.show()

def find_minima(index, tic, right=True):
            
    j = index
    #apex_abundance = tic[index]
    tic_len = len(tic)

    if right: minima = tic[j] >= tic[j+1]
    else: minima = tic[j] >= tic[j-1]

    while minima:
        
        if j == 1 or j == tic_len -2:
            break
        
        if right: 
            j += 1

            minima = tic[j] >= tic[j+1]

        else: 
            j -= 1
            minima = tic[j] >= tic[j-1]

    if right: return j
    else: return j

def derivate(data_array):

    data_array = np.array(data_array)

    dy = data_array[1:] - data_array[:-1]

    '''replaces nan for infinity'''
    indices_nan = np.where(np.isnan(data_array))[0]

    if indices_nan.size:

        data_array[indices_nan] = np.inf
        dy[np.where(np.isnan(dy))[0]] = np.inf

    return dy

def minima_detector(tic, max_tic, peak_height_max_percent, peak_max_prominence_percent):

    peak_height_diff = lambda hi, li : ((tic[hi] - tic[li]) / max_tic )*100

    for start_index, index, final_index in peak_detector(tic, max_tic):

        # abundance max threshold    
        if (tic[index] / max_tic) * 100 < peak_height_max_percent:

            # calculates prominence and filter   
            if peak_height_diff(index, start_index) and peak_height_diff(index, final_index) < peak_max_prominence_percent:
                
                    yield from (start_index, final_index)

def baseline_detector(tic, rt, peak_height_max_percent, peak_max_prominence_percent, do_interpolation=True):

    rt = np.array(rt)

    max_tic = max(tic)

    indexes = sorted(list(set(i for i in minima_detector(tic, max_tic, peak_height_max_percent, peak_max_prominence_percent))))

    y = -tic

    x1 = rt[indexes]

    y1 = y[indexes]

    if len(x1) <= 5:
        return tic

    if not do_interpolation:
  
        y0 = np.zeros(tic.shape)
        y0[indexes] = y[indexes]

        return y0

    else:

        f1 = interpolate.interp1d(x1, y1, kind='quadratic',fill_value="extrapolate")

        ynew1 = f1(list(rt))

        # from matplotlib import pyplot as plt   
        # if self.deconv_rt_list and  self.deconv_mz == 51:

        #   plt.plot(rt, tic-(-1* ynew1), color='green')

        # plt.plot(rt, -1* ynew1, c='black')

        # s = self.smooth(s, 10, 'blackman')

        # plt.plot(self.retention_time, -s)

        # plt.show()

        return -1 * ynew1

def peak_detector_generator(tic, stds, method, rt, max_height, min_height, max_prominence, min_datapoints):

    max_tic = max(tic)

    if method == 'manual_relative_abundance':

        tic = tic - baseline_detector(tic, rt, max_height, max_prominence)

        norm_tic = (tic / max_tic) * 100

        remove_indexes = np.where(norm_tic < min_height)[0]

        # if self.deconv_rt_list and  self.deconv_mz == 51:
        #    plt.plot(self.deconv_rt_list, tic, label=self.deconv_mz)

    elif method == 'auto_relative_abundance':

        tic = tic - baseline_detector(tic, rt, max_height, max_prominence)

        baseline = baseline_detector(tic, rt, max_height, max_prominence)

        peak_detect_threshold = ((np.nanmean(baseline) + (stds * np.std(baseline))))

        remove_indexes = np.where(tic < peak_detect_threshold)[0]

    elif method == 'second_derivative':

        remove_indexes = second_derivative_threshold(tic, stds, rt, max_height, max_prominence)

    else:

        NotImplemented(method)

    peak_height_diff = lambda hi, li : ((tic[hi] - tic[li]) / max_tic )*100
    
    dy = derivate(tic)

    include_indexes = np.where((np.hstack((dy, 0)) < 0) & (np.hstack((0, dy)) > 0))[0]

    final_indexes = sorted(set(include_indexes)-set(remove_indexes))

    #from matplotlib import pyplot as plt   
    
    #plt.plot(self.retention_time, tic, color='black')
    #plt.scatter(self.retention_time[remove_indexes], tic[remove_indexes], color='red')
    #plt.scatter(self.retention_time[include_indexes], tic[include_indexes], color='blue')
    #plt.scatter(self.retention_time[final_indexes], tic[final_indexes], color='blue')
    
    #plt.show()

    for index in final_indexes:
            
        start_index = find_minima(index, tic, right=False)
        final_index = find_minima(index, tic)
        
        if final_index-start_index > min_datapoints:

            #if min( peak_height_diff(index,start_index), peak_height_diff(index,final_index) )> self.chromatogram_settings.peak_min_prominence_percent :   
                
                yield (start_index, index, final_index)

def smooth_signal(x, window_len, window, pol_order, implemented_smooth_method):
    
    """smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal 
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.

    output:
        the smoothed signal
        
    see also: 

    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.savgol_filter

    """
    x= np.array(x)

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input array needs to be bigger than window size")

    #if window_len < 3:
    #    return x

    if not window in implemented_smooth_method:
        raise ValueError("Window method should be 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")

    s = np.r_[x[window_len - 1:0:-1], x, x[-1:-window_len:-1]]

    if window == 'savgol':
        
        return savgol_filter(x, window_len, pol_order)


    elif window == 'boxcar':  # moving average
        
        w = boxcar(window_len)
        
        y = np.convolve(w, s, mode='valid')

        
    elif window == 'flat':  # moving average
        
        w = np.ones(window_len, 'd')

        y = np.convolve(w / w.sum(), s, mode='valid')
        
    else:
        
        w = eval( window + '(window_len)')

        y = np.convolve(w / w.sum(), s, mode='valid')

    return y[int(window_len / 2 - 1):int(-window_len / 2)]

def second_derivative_threshold(tic, stds, rt, peak_height_max_percent, peak_max_prominence_percent):
          
    dy = derivate(tic)
    
    dydy = derivate(dy)
    dydy = np.hstack((dydy, 0))
    dydy = np.hstack((0, dydy))

    baseline = baseline_detector(dydy, rt, peak_height_max_percent, peak_max_prominence_percent, do_interpolation=False)
    
    threshold_median = np.median(baseline) - (stds * np.std(baseline))
    
    remove_indexes = np.where(dydy > threshold_median)[0]
    
    return remove_indexes
        


__author__ = "Yuri E. Corilo"
__date__ = "Jun 24, 2019"

from IsoSpecPy import IsoSpecPy
from numpy import isnan, power, exp, nextafter
from pandas import DataFrame
from scipy.stats import pearsonr, spearmanr, kendalltau

from corems.encapsulation.constant import Atoms
from corems.encapsulation.constant import Labels
from corems.encapsulation.factory.parameters import MSParameters
from corems.molecular_id.calc.SpectralSimilarity import SpectralSimilarity

class MolecularFormulaCalc:
    
    def _calc_resolving_power_low_pressure(self, B, T):
        '''
        ## Parameters
        ----------
        #### T << collisional damping time
        #### T = transient time (seconds)
        #### B = Magnetic Strength (Testa)
        '''
        return (1.274 * 10000000 * B * T) *(1/self.mz_calc)    

    def _calc_resolving_power_high_pressure(self, B, T):
        '''
        ## Parameters
        ----------
        #### T << collisional damping time 
        #### t= collisional dumping constant
        #### T = transient time (seconds)
        #### B = Magnetic Strength (Testa)
        '''
        return (2.758 * 10000000 * B * T) *(1/self.mz_calc)    

    def _calc_mz(self):
        
        def protonated_mass(mass, ion_charge):
            return (mass + (ion_charge * Atoms.atomic_masses.get("H")) + (ion_charge * -1 * Atoms.electron_mass))/abs(ion_charge)
        
        def radical_mass(mass, ion_charge):
            return (mass + (ion_charge * -1 * Atoms.electron_mass))/ abs(ion_charge)

        if self.ion_charge:
            
            mass = 0
            
            ion_type = self._d_molecular_formula.get(Labels.ion_type)

            for each_atom in self._d_molecular_formula.keys() :
                
                if each_atom != Labels.ion_type and each_atom != 'HC':
                    
                    try:
                        mass = mass + Atoms.atomic_masses[each_atom]  *  self._d_molecular_formula.get(each_atom)
                    except: print(Labels.ion_type, each_atom) 
            
            if ion_type == Labels.protonated_de_ion:
                return protonated_mass(mass, self.ion_charge)
            elif ion_type == Labels.radical_ion or ion_type == Labels.adduct_ion:   
                return radical_mass(mass, self.ion_charge)
            else:
                #formula is probably ion form used for bruker ref list
                return radical_mass(mass, self.ion_charge)
                
        else:
            
            raise Exception("Please set ion charge first")
         
    def _calc_assignment_mass_error(self, method='ppm'):
        
        '''
        ## Parameters
        ----------
        #### mz_exp: float, 
        ####       Experimental m/z 
        #### method: string, 
        ####       ppm or ppb
        '''
         
        if method == 'ppm':
            multi_factor = 1000000
        
        elif method == 'ppb':
            multi_factor = 1000000
        
        else:
            raise Exception("method needs to be ppm or ppb, you have entered %s" % method)
              
        if self._mspeak_parent.mz_exp:
            
            self._assignment_mass_error = ((self.mz_calc - self._mspeak_parent.mz_exp)/self.mz_calc)*multi_factor

            return ((self.mz_calc - self._mspeak_parent.mz_exp)/self.mz_calc)*multi_factor
        
        else:
            
            raise Exception("No ms peak associated with the molecular formula instance %s", self)    
    
    
    def _calc_mz_confidence(self, mean=0):
        
        # predicted std not set, using 0.3
        if not self._mspeak_parent.predicted_std: self._mspeak_parent.predicted_std = 0.3
        
        #print( self._mspeak_parent.predicted_std)
        
        return  exp( -1 * (power((self.mz_error -  mean),2)  / (2 * power(self._mspeak_parent.predicted_std,2)) ))
    
    def _calc_isotopologue_confidence(self):

        if self.is_isotopologue:
            # confidence of isotopologue is pure mz error 
            # TODO add more features here 
            
            mformula_index = self.mono_isotopic_formula_index
            mspeak_index = self.mspeak_index_mono_isotopic

            mspeak = self._mspeak_parent._ms_parent[mspeak_index]
            
            expected_isotopologues = mspeak[mformula_index].expected_isotopologues
            
            mono_mz = mspeak[mformula_index].mz_calc
            mono_abundance = mspeak.abundance

        else:

            mono_mz = self.mz_calc
            mono_abundance = self._mspeak_parent.abundance

            expected_isotopologues = self.expected_isotopologues
            # has isotopologues based on current dinamic range
            
        if expected_isotopologues:
            
            dict_mz_abund_ref = {'mz':[mono_mz], 'abundance':[mono_abundance]}
            
            # get reference data
            for mf in expected_isotopologues:
                dict_mz_abund_ref['abundance'].append(mf.abundance_calc)
                dict_mz_abund_ref['mz'].append(mf.mz_calc)

            dict_mz_abund_exp = {mono_mz:mono_abundance}
            
            # get experimental data
            for mf in expected_isotopologues:
                
                # molecular formula has been assigned to a peak
                if mf._mspeak_parent:
                    #stores mspeak abundance
                    dict_mz_abund_exp[mf.mz_calc] = mf._mspeak_parent.abundance
                    
                else:
                    # fill missing mz with abundance 0 and mz error score of 0
                    dict_mz_abund_exp[mf.mz_calc] = nextafter(0, 1)
            
            distance = SpectralSimilarity(dict_mz_abund_exp, dict_mz_abund_ref).manhattan_distance()
            correlation = 1 - self.normalize_distance(distance, [0, 2])
            #correlation = dwt_correlation(dict_mz_abund_exp, dict_mz_abund_ref)
            #correlation = cosine_correlation(dict_mz_abund_exp, dict_mz_abund_ref)
            
            if correlation == 1:
                print(dict_mz_abund_exp,dict_mz_abund_ref)
            if isnan(correlation):
                #print(dict_mz_abund_exp,dict_mz_abund_ref)
                correlation = 0.00001
        
        else:
            
            # no isotopologue expected giving a correlation score of 0.0 but it needs optimization
            correlation = 0.0

        return correlation

    def normalize_distance(self, dist, dist_range):
    
        result = (dist - dist_range[0]) / (dist_range[1] - dist_range[0])

        if result < 0:
            result = 0.
        elif result > 1:
            result = 1.

        return result

    def _calc_average_mz_score(self):
        
        if self.is_isotopologue:
            # confidence of isotopologue is pure mz error 
            # TODO add more features here 
            
            mformula_index = self.mono_isotopic_formula_index
            mspeak_index = self.mspeak_index_mono_isotopic

            mspeak = self._mspeak_parent._ms_parent[mspeak_index]
            
            expected_isotopologues = mspeak[mformula_index].expected_isotopologues

        else:
            
            expected_isotopologues = self.expected_isotopologues
            # has isotopologues based on current dinamic range
        
        accumulated_mz_score = [self.mz_error_score]
        
        if expected_isotopologues:
            
            for mf in expected_isotopologues:
                # molecular formula has been assigned to a peak
                if mf._mspeak_parent:
                    #stores mspeak abundance
                    accumulated_mz_score.append(mf.mz_error_score)
                else:
                    # fill missing mz with abundance 0 and mz error score of 0
                    accumulated_mz_score.append(0.0)

        average_mz_score = sum(accumulated_mz_score)/len(accumulated_mz_score)
        
        if isnan(average_mz_score):
                average_mz_score = 0.0

        return average_mz_score       

    def _calc_confidence_score(self):
        
        '''
        ### Assumes random mass error, i.e, spectrum has to be calibrated and with zero mean
        #### TODO: Add spectral similarity 

        ## Parameters
        ----------
        #### mz_exp:
        ####    Experimental m/z 
        #### predicted_std:
        ####    Standart deviation calculated from Resolving power optimization or constant set by User 
        
        '''
        
        isotopologue_correlation = self.isotopologue_similarity
        average_mz_score = self.average_mz_error_score
        # add monoisotopic peak mz error score
        
        # calculate score with higher weight for mass error
        #score = power(((isotopologue_correlation) * (power(average_mz_score,3))),1/4)
        a = self._mspeak_parent._ms_parent.molecular_search_settings.mz_error_score_weight
        b = self._mspeak_parent._ms_parent.molecular_search_settings.isotopologue_score_weight
        
        score = (isotopologue_correlation*b) + (average_mz_score*a)
        
        #if round(average_mz_score,2) == 0.00:
        #    print(a,b, average_mz_score, isotopologue_correlation, score, isotopologue_correlation*b)
        

        return score


    def _calc_abundance_error(self, method='percentile'):
        '''method should be ppm, ppb or percentile'''
        
        mult_factor = 100

        iso_abundance = self._mspeak_parent.abundance
        mono_abundance =self._mspeak_parent._ms_parent[self.mspeak_index_mono_isotopic].abundance

        if self.prob_ratio:
            
            theor_abundance = mono_abundance* self.prob_ratio
            #self.parent need to have a MassSpecPeak associated with the MolecularFormula class
            return ((theor_abundance - iso_abundance )/theor_abundance)*mult_factor
        
        else:
            
            raise Exception("Please calc_isotopologues")    

    def _calc_area_error(self, method='percentile'):
        '''method should be ppm, ppb or percentile'''
        
        mult_factor = 100
        
        iso_area = self._mspeak_parent.area
        mono_area =self._mspeak_parent._ms_parent[self.mspeak_index_mono_isotopic].area

        if self.prob_ratio:
            
            if mono_area and iso_area: 

                #exp_ratio = iso_area/mono_area  
                          
                area_calc = mono_area* self.prob_ratio

                #self.parent need to have a MassSpecPeak associated with the MolecularFormula class
                return ((area_calc - iso_area )/area_calc)*mult_factor
                #return ((self.prob_ratio - exp_ratio )/self.prob_ratio)*mult_factor
            
            else:
                
                #centroid mass spectrum
                return 0
        else:
            
            raise Exception("Please calc_isotopologues")    

    @property
    def dbe_ai(self):
            
        carbons =  self._d_molecular_formula.get('C')
        hydrogens = self._d_molecular_formula.get('H')
        oxygens = self._d_molecular_formula.get('O')
        return 1 + (((2*carbons) - hydrogens - (2*oxygens))*0.5)

    def _calc_dbe(self):
            
            individual_dbe = 0
            
            for atom in self._d_molecular_formula.keys():
                
                if atom != Labels.ion_type:
                    
                    n_atom = int(self._d_molecular_formula.get(atom))
                    
                    clean_atom = ''.join([i for i in atom if not i.isdigit()]) 
                    
                    if self._mspeak_parent:
                        valencia = self._mspeak_parent._ms_parent.molecular_search_settings.used_atom_valences.get(clean_atom)
                    else:
                        valencia = MSParameters.molecular_search.used_atom_valences.get(clean_atom)
                    #valencia = Atoms.atoms_covalence.get(atom)
                    
                    if type(valencia) is tuple:
                        valencia = valencia[0]
                    if valencia > 0:
                        #print atom, valencia, n_atom, individual_dbe
                        individual_dbe = individual_dbe + (n_atom * (valencia - 2))
                    else:
                        continue
            
            dbe = 1 + (0.5 * individual_dbe)
            
            
            
            if self.ion_type == Labels.adduct_ion:
                dbe = dbe + 0.5
            
            return dbe

    def _calc_kdm(self, dict_base):
        '''dict_base = {"C": 1, "H": 2}
        '''
        mass = 0
        for atom in dict_base.keys():
            mass = mass + Atoms.atomic_masses.get(atom) * dict_base.get(atom)
        
        kendrick_mass = (int(mass)/mass)*self.mz_calc
        
        nominal_km =int(kendrick_mass)
       
        kmd = (nominal_km - kendrick_mass) * 100
        
        #kmd = (nominal_km - km) * 1
        kdm  = round(kmd,0)
        
        return kdm, kendrick_mass, nominal_km

    def _cal_isotopologues(self, formula_dict, min_abundance, current_abundance, ms_dynamic_range):
        
        '''
        primary function to look for isotopologues based on a monoisotopic molecular formula
        INPUT {'C':10, 'H', 20, 'O', 2, etc} Atomic labels need to follow Atoms class atoms labels
        
        This function needs to be expanded to include the calculation of resolving power
        and plot the results.
        
        *   use this function at runtime during the molecular identification algorithm
            only when a positive ID is observed to the monoisotopic ion
        
        *   use this function to simulate mass spectrum 
            (needs resolving power calculation to be fully operational)        
            last update on 05-26-2020, Yuri E. Corilo 
        
        *   it might break when adding non-conventional atoms (not yet tested)
            
        *   it needs speed optimization; update: (Using IsoSpeccPy, a C Library (fast and accurate)) 
            https://github.com/MatteoLacki/IsoSpec
        '''
        # updated it to reflect min possible mass peak abundance
        cut_off_to_IsoSpeccPy = 1-(1/ms_dynamic_range)
        
        #print("cut_off_to_IsoSpeccPy", cut_off_to_IsoSpeccPy, current_abundance, min_abundance, ms_dynamic_range)
        #print(cut_off_to_IsoSpeccPy)
        atoms_labels = (atom for atom in formula_dict.keys() if atom != Labels.ion_type and atom != 'H')
       
        atoms_count = []
        masses_list_tuples = []
        props_list_tuples = []
        all_atoms_list = []
        
        for atom_label in atoms_labels:
            
            if not len(Atoms.isotopes.get(atom_label))>1:
                'This atom_label has no heavy isotope'
                atoms_count.append(formula_dict.get(atom_label))
                mass = Atoms.atomic_masses.get(atom_label)
                prop = Atoms.isotopic_abundance.get(atom_label)
                masses_list_tuples.append([mass])
                props_list_tuples.append([prop])
                all_atoms_list.append(atom_label)
                
            else:
                
                isotopes_label_list = Atoms.isotopes.get(atom_label)[1]
            
                if len(isotopes_label_list) > 1:
                    'This atom_label has two or more heavy isotope'
                    isotopos_labels = [i for i in isotopes_label_list]
                else:
                    'This atom_label only has one heavy isotope'
                    isotopos_labels = [isotopes_label_list[0]]
                
                #all_atoms_list.extend(isotopos_labels) 
                isotopos_labels = [atom_label] + isotopos_labels
                
                all_atoms_list.extend(isotopos_labels)
                
                masses = [Atoms.atomic_masses.get(atom_label) for atom_label in isotopos_labels]
                props = [Atoms.isotopic_abundance.get(atom_label) for atom_label in isotopos_labels]
                
                atoms_count.append(formula_dict.get(atom_label))
                masses_list_tuples.append(masses)
                props_list_tuples.append(props)
        
        iso = IsoSpecPy.IsoSpec(atoms_count,masses_list_tuples,props_list_tuples, cut_off_to_IsoSpeccPy)
        
        conf = iso.getConfs()
        masses = conf[0]
        probs = exp(conf[1])
        molecular_formulas = conf[2]
        #print('conf', conf)
        #print('probs', conf[1])
        
        new_formulas = []
        
        for isotopologue_index in range(len(iso)):
            #skip_mono_isotopic 
            
            formula_list = molecular_formulas[isotopologue_index]
            new_formula_dict = dict(zip(all_atoms_list, formula_list))
            new_formula_dict[Labels.ion_type] = formula_dict.get(Labels.ion_type)
            if formula_dict.get('H'):
                new_formula_dict['H'] = formula_dict.get('H')

            new_formulas.append({x:y for x,y in new_formula_dict.items() if y!=0})
        
        # formula_dict in new_formulas check if monoisotopic is being returned
        if new_formulas:# and formula_dict in new_formulas:
            
            #print(conf)    
            #print(new_formulas)    
            #print(atoms_count)
            #print(all_atoms_list)
            #print(masses_list_tuples)
            #print(props_list_tuples)
            # find where monoisotopic is
            index_mono = new_formulas.index(formula_dict)   
            # calculate ratio iso/mono
            probs = list(probs/probs[index_mono])
            
            # delete the monoisotopic
            del probs[index_mono]
            del new_formulas[index_mono]
            
            #print('probs_exp', probs)
            for formulas, prob in zip(new_formulas, probs):
                
                theor_abundance = current_abundance* prob
                if theor_abundance > min_abundance:
                    #print(prob, theor_abundance, current_abundance)
                    yield (formulas, prob)
            #return zip(new_formulas, probs )
    
        #else:
        #    return []    
    


__author__ = "Yuri E. Corilo"
__date__ = "Feb 13, 2020"


from collections.abc import Mapping
from pathlib import Path
import json

from numpy import array

from corems.mass_spectra.calc.GC_Calc import GC_Calculations
from corems.mass_spectra.calc.GC_Deconvolution import MassDeconvolution
from corems.mass_spectra.calc import SignalProcessing as sp

from corems.chroma_peak.factory.ChromaPeakClasses import GCPeak
from corems.mass_spectra.output.export import LowResGCMSExport
from corems.encapsulation.factory.parameters import GCMSParameters


class GCMSBase(GC_Calculations, MassDeconvolution):
    """
    classdocs
    """

    def __init__(self, file_location, analyzer='Unknown', instrument_label='Unknown', sample_name=None):

        """
            # Parameters
        ----------
        file_location: text,  pathlib.Path(), or s3path.S3Path
            Path object from pathlib containing the file location
        """
        if isinstance(file_location, str):
            # if obj is a string it defaults to create a Path obj, pass the S3Path if needed
            file_location = Path(file_location)

        if not file_location.exists():

            raise FileExistsError("File does not exist: " + str(file_location))

        self.file_location = file_location

        if sample_name: self.sample_name = sample_name
        else: self.sample_name = file_location.stem

        self.analyzer = analyzer
        self.instrument_label = instrument_label
        self._init_settings()

        self._retention_time_list = []
        self._scans_number_list = []
        self._tic_list = []

        # all scans
        self._ms = {}

        # after peak detection
        self._processed_tic = []
        self.gcpeaks = []

        self.ri_pairs_ref = None
        self.cal_file_path = None

    def _init_settings(self):

        self._parameters = GCMSParameters()

    def __len__(self):

        return len(self.gcpeaks)

    def __getitem__(self, scan_number):

        return self.gcpeaks[scan_number]

    # def __iter__(self):

    #     return iter(self.gcpeaks.values())

    def process_chromatogram(self, plot_res=False):

        # tic = self.tic - self.baseline_detector(self.tic)

        self._processed_tic = self.smooth_tic(self.tic)

        for index, tic in enumerate(self._processed_tic):

            self._ms[index]._processed_tic = tic

        # self.second_derivative_threshold(self._processed_tic)

        if self.chromatogram_settings.use_deconvolution:

            self.run_deconvolution(plot_res=False)

        else:

            peaks_index = self.centroid_detector(self._processed_tic, self.retention_time)

            for i in peaks_index:

                apex_index = i[1]

                gc_peak = GCPeak(self, self._ms[apex_index], i )

                gc_peak.calc_area(self._processed_tic, 1)

                self.gcpeaks.append(gc_peak)

                # self.gcpeaks[self.scans_number[apex_index]] = gc_peak

    def add_mass_spectrum(self, mass_spec):

        self._ms[mass_spec.scan_number] = mass_spec

    def set_tic_list_from_data(self):

        self.tic = [self._ms.get(i).tic for i in self.scans_number]

        # self.set_tic_list([self._ms.get(i).get_sumed_signal_to_noise() for i in self.get_scans_number()])

    def set_retention_time_from_data(self):

        retention_time_list = []

        for key_ms in sorted(self._ms.keys()):

            retention_time_list.append(self._ms.get(key_ms).rt)

        self.retention_time = retention_time_list

        # self.set_retention_time_list(sorted(self._ms.keys()))

    def set_scans_number_from_data(self):

        self.scans_number = sorted(self._ms.keys())

    @property
    def parameter(self):
        return self._parameters

    @parameter.setter
    def parameter(self, gcms_parameters_instance):
        self._parameters = gcms_parameters_instance

    @property
    def molecular_search_settings(self):
        return self.parameter.molecular_search

    @molecular_search_settings.setter
    def molecular_search_settings(self, settings_class_instance):

        self.parameter.molecular_search = settings_class_instance

    @property
    def chromatogram_settings(self):
        return self.parameter.gc_ms

    @chromatogram_settings.setter
    def chromatogram_settings(self, settings_class_instance):

        self.parameter.gc_ms = settings_class_instance

    @property
    def scans_number(self):

        return self._scans_number_list

    @property
    def retention_time(self):

        return self._retention_time_list

    @property
    def processed_tic(self):

        return self._processed_tic

    @property
    def tic(self):

        return self._tic_list

    @property
    def max_tic(self):

        return max([gc_peak.tic for gc_peak in self])

    @property
    def min_tic(self):

        return min([gc_peak.tic for gc_peak in self])

    @property
    def dynamic_range(self):

        return self.max_tic / self.min_tic

    @property
    def matched_peaks(self):
        return [gc_peak for gc_peak in self if gc_peak]

    @property
    def sorted_gcpeaks(self):
        return sorted(self, key=lambda g: g.rt)

    @property
    def unique_metabolites(self):

        metabolites = set()
        for gc_peak in self:
            if gc_peak:
                for compound_obj in gc_peak:
                    metabolites.add(compound_obj.name)

        return metabolites

    @property
    def no_matched_peaks(self):
        return [peak for peak in self if not peak]

    @retention_time.setter
    def retention_time(self, l):
        # self._retention_time_list = linspace(0, 80, num=len(self._scans_number_list))
        self._retention_time_list = l

    @scans_number.setter
    def scans_number(self, l):

        self._scans_number_list = l

    @tic.setter
    def tic(self,l):

        self._tic_list = array(l)    

    def plot_gc_peaks(self, ax=None, color="red"):  # pragma: no cover

        import matplotlib.pyplot as plt
        fig = plt.gcf()
        if ax is None:
            ax = plt.gca()

        max_rts = [gc_peak.mass_spectrum.rt for gc_peak in self]
        max_tics = [gc_peak.mass_spectrum.tic for gc_peak in self]

        # min_rts = [self._ms[gc_peak.start_index].rt for gc_peak in self] + [self._ms[gc_peak.final_index].rt for gc_peak in self]
        # min_tics = [self._ms[gc_peak.start_index].tic for gc_peak in self] + [self._ms[gc_peak.final_index].tic for gc_peak in self]
        # sc = ax.scatter(min_rts, min_tics, color='yellow', linewidth=0, marker='v')

        sc = ax.scatter(max_rts, max_tics, color=color, marker='v')

        ax.set(xlabel='Retention Time (s)', ylabel='Total Ion Chromatogram')

        annot = ax.annotate("", xy=(0, 0), xytext=(20, 20), textcoords="offset points",
                            bbox=dict(boxstyle="round", fc="w"),
                            arrowprops=dict(arrowstyle="->"))
        annot.set_visible(False)
        annot.get_bbox_patch().set_facecolor(('lightblue'))
        annot.get_bbox_patch().set_alpha(0.8)

        def update_annot(ind):

            pos = sc.get_offsets()[ind["ind"][0]]
            annot.xy = pos

            text = "RT: {}\nRT Ref: {}\nRI: {}\nRI Ref: {}\nSimilarity Score: {}\nName: {}".format(" ".join([str(round(self[n].rt, 2)) for n in ind["ind"]]),
                           " ".join([str(round(self[n].highest_score_compound.rt, 2) if self[n].highest_score_compound else None) for n in ind["ind"]]),
                           " ".join([str(round(self[n].ri, 2) if self[n].ri else None) for n in ind["ind"]]),
                           " ".join([str(round(self[n].highest_score_compound.ri, 2) if self[n].highest_score_compound else None) for n in ind["ind"]]),                           
                           " ".join([str(round(self[n].highest_score_compound.similarity_score, 4) if self[n].highest_score_compound else None) for n in ind["ind"]]),
                           " ".join([str(self[n].highest_score_compound.name if self[n].highest_score_compound else None) for n in ind["ind"]])
                           )
            annot.set_text(text)

        def hover(event):
            vis = annot.get_visible()
            if event.inaxes == ax:
                cont, ind = sc.contains(event)
                if cont:
                    update_annot(ind)
                    annot.set_visible(True)
                    fig.canvas.draw_idle()
                else:
                    if vis:
                        annot.set_visible(False)
                        fig.canvas.draw_idle()

        fig.canvas.mpl_connect("motion_notify_event", hover)

        return ax

    def to_excel(self, out_file_path, write_mode='ab', id_label="corems:"):

        exportMS = LowResGCMSExport(out_file_path, self)
        exportMS.to_excel(id_label=id_label)

    def to_csv(self, out_file_path, write_mode='ab', id_label="corems:"):

        exportMS = LowResGCMSExport(out_file_path, self)
        exportMS.to_csv(id_label=id_label)

    def to_pandas(self, out_file_path, id_label="corems:"):

        # pickle dataframe (pkl extension)
        exportMS = LowResGCMSExport(out_file_path, self)
        exportMS.to_pandas(id_label=id_label)

    def to_dataframe(self, id_label="corems:"):

        # returns pandas dataframe
        exportMS = LowResGCMSExport(self.sample_name, self)
        return exportMS.get_pandas_df(id_label=id_label)

    def processing_stats(self):

        # returns json string
        exportMS = LowResGCMSExport(self.sample_name, self)
        return exportMS.get_data_stats(self)

    def parameters_json(self, id_label="corems:", output_path=" "):

        # returns json string
        exportMS = LowResGCMSExport(self.sample_name, self)
        return exportMS.get_parameters_json(self, id_label, output_path)

    def to_json(self, id_label="corems:"):

        # returns pandas dataframe
        exportMS = LowResGCMSExport(self.sample_name, self)
        return exportMS.get_json(id_label=id_label)

    def to_hdf(self, id_label="corems:"):

        # returns pandas dataframe
        exportMS = LowResGCMSExport(self.sample_name, self)
        return exportMS.to_hdf(id_label=id_label)

    def plot_chromatogram(self, ax=None, color="blue"):  # pragma: no cover

        import matplotlib.pyplot as plt

        if ax is None:
            ax = plt.gca()

        ax.plot(self.retention_time, self.tic, color=color)
        ax.set(xlabel='Retention Time (s)', ylabel='Total Ion Chromatogram')

        return ax

    def plot_smoothed_chromatogram(self, ax=None, color="green"):  # pragma: no cover

        import matplotlib.pyplot as plt

        if ax is None:

            ax = plt.gca()

        ax.plot(self.retention_time, self.smooth_tic(self.tic), color=color)

        ax.set(xlabel='Retention Time (s)', ylabel='Total Ion Chromatogram')

        return ax

    def plot_detected_baseline(self, ax=None, color="blue"):  # pragma: no cover

        import matplotlib.pyplot as plt

        if ax is None:

            ax = plt.gca()

        max_height = self.chromatogram_settings.peak_height_max_percent
        max_prominence = self.chromatogram_settings.peak_max_prominence_percent

        baseline = sp.baseline_detector(self.tic, self.retention_time, max_height, max_prominence)
        ax.plot(self.retention_time, color=color)
        ax.set(xlabel='Retention Time (s)', ylabel='Total Ion Chromatogram')

        return ax

    def plot_baseline_subtraction(self, ax=None, color="black"):  # pragma: no cover

        import matplotlib.pyplot as plt

        if ax is None:

            ax = plt.gca()

        max_height = self.chromatogram_settings.peak_height_max_percent

        max_prominence = self.chromatogram_settings.peak_max_prominence_percent

        x = self.tic + sp.baseline_detector(self.tic, self.retention_time, max_height, max_prominence)

        ax.plot(self.retention_time, x, color=color)

        ax.set(xlabel='Retention Time (s)', ylabel='Total Ion Chromatogram')

        return ax

    def peaks_rt_tic(self, json_string=False):

        peaks_list = dict()

        all_candidates_data = {}

        all_peaks_data = {}

        for gcms_peak in self.sorted_gcpeaks:

            dict_data = {'rt': gcms_peak.rt_list,
                         'tic': gcms_peak.tic_list,
                         'mz': gcms_peak.mass_spectrum.mz_exp.tolist(),
                         'abundance': gcms_peak.mass_spectrum.abundance.tolist(),
                         'candidate_names': gcms_peak.compound_names,
                         }

            peaks_list[gcms_peak.rt] = dict_data

            for compound in gcms_peak:

                if compound.name not in all_candidates_data.keys():
                    mz = array(compound.mz).tolist()
                    abundance = array(compound.abundance).tolist()
                    data = {'mz': mz, "abundance": abundance}
                    all_candidates_data[compound.name] = data

        all_peaks_data["peak_data"] = peaks_list
        all_peaks_data["ref_data"] = all_candidates_data

        if json_string:

            return json.dumps(all_peaks_data)

        else:
            return all_peaks_data

    def plot_processed_chromatogram(self, ax=None, color="black"):

        import matplotlib.pyplot as plt

        if ax is None:

            ax = plt.gca()

        ax.plot(self.retention_time, self.processed_tic, color=color)

        ax.set(xlabel='Retention Time (s)', ylabel='Total Ion Chromatogram')

        return ax



__author__ = "Yuri E. Corilo"
__date__ = "Jun 12, 2019"
from copy import deepcopy
from pathlib import Path

from numpy import genfromtxt, fromstring, dtype, fromfile, frombuffer
from s3path import S3Path
from xml.dom import minidom


from corems.transient.factory.TransientClasses import Transient
from corems.encapsulation.factory.parameters import default_parameters



class ReadBrukerSolarix(object):
    
    """
    A class used to Read a single Transient from Bruker's FT-MS acquisition station (fid, or ser)
        
    Parameters
    ----------
    d_directory_location : str
        the full path of the .d folder
    
    Methods
    -------
    get_transient()
        Read the data and settings returning a Transient class  
    """
    
    def __enter__(self ):
            
            return self.get_transient()
    
    def __exit__(self, exc_type, exc_val, exc_tb):
            
        return False

    def __init__(self, d_directory_location):
        
        if isinstance(d_directory_location, str):
            d_directory_location = Path(d_directory_location)
        
        if not d_directory_location.exists():
            raise FileNotFoundError("File does not exist: " + str(d_directory_location))

        self.d_directory_location = d_directory_location
        
        self.file_location = d_directory_location
        
        try:

            self.parameter_filename_location = self.locate_file(
                d_directory_location, "apexAcquisition.method"
            )
            self.transient_data_path = d_directory_location / "fid"
            
            if not self.transient_data_path.exists():

                self.transient_data_path = d_directory_location / "ser"

                if not self.transient_data_path.exists():
                    
                    raise FileNotFoundError("Could not locate transient data")

                else:
                    # get scan attributes
                    self.scan_attr = d_directory_location / "scan.xml"

        except:
            
            raise FileExistsError(
                "%s does not seem to be a valid Solarix Mass Spectrum"
                % (d_directory_location)
            )

    def get_scan_attr(self):
    
        from bs4 import BeautifulSoup
        
        soup = BeautifulSoup(self.scan_attr.open(),'xml')

        list_rt = [float(rt.text) for rt in soup.find_all('minutes')]
        list_tic = [float(tic.text) for tic in soup.find_all('tic')]
        list_scan = [int(scan.text) for scan in soup.find_all('count')]

        dict_scan_rt_tic = dict(zip(list_scan, zip(list_rt, list_tic)))
        
        return dict_scan_rt_tic
       
        
    def get_transient(self, scan_number=1):

        file_d_params = self.parse_parameters(self.parameter_filename_location)

        self.fix_freq_limits(file_d_params)

        from sys import platform
        
        if platform == "win32":
            # Windows...
            dt = dtype("l")
        else:
            dt = dtype("i")

        # get rt, scan, and tic from scan.xml file, otherwise  using 0 defaults values 
        
        output_parameters = deepcopy(default_parameters(self.d_directory_location))

        if self.transient_data_path.name == 'ser':
            
            if self.scan_attr.exists():
                
                dict_scan_rt_tic = self.get_scan_attr()

                output_parameters["scan_number"] = scan_number

                output_parameters["rt"] = dict_scan_rt_tic.get(scan_number)[0]

                output_parameters["tic"] = dict_scan_rt_tic.get(scan_number)[1]
        
        output_parameters["analyzer"] = "ICR"

        output_parameters["label"] = "Bruker_Frequency"

        output_parameters["Aterm"] = float(file_d_params.get("ML1"))

        output_parameters["Bterm"] = float(file_d_params.get("ML2"))

        output_parameters["Cterm"] = float(file_d_params.get("ML3"))

        output_parameters["exc_high_freq"] = float(file_d_params.get("EXC_Freq_High"))

        output_parameters["exc_low_freq"] = float(file_d_params.get("EXC_Freq_Low"))

        output_parameters["bandwidth"] = float(file_d_params.get("SW_h"))

        output_parameters["number_data_points"] = int(file_d_params.get("TD"))

        output_parameters["polarity"] = str(file_d_params.get("Polarity"))

        data_points = int(file_d_params.get("TD"))

        scan = output_parameters["scan_number"]
        from io import BytesIO
        if self.transient_data_path.name == 'ser':
            
            if isinstance(self.transient_data_path, S3Path):
                databin = BytesIO(self.transient_data_path.open('rb').read())
            
            else:
                databin = self.transient_data_path.open('rb')
               
            databin.seek((scan-1)*4*data_points)
            #read scan data and parse to 32int struct
            data = frombuffer(databin.read(4*data_points), dtype=dt)
        
        else:
            
            if isinstance(self.transient_data_path, S3Path):
                data = frombuffer(self.transient_data_path.open('rb').read(), dtype=dt)
            else:
                data = fromfile(self.transient_data_path, dtype=dt)
        
        return Transient(data, output_parameters)

    """
        for key, values in default_parameters.items():
            print(key, values)
    """
    def fix_freq_limits(self, d_parameters):

        highfreq = float(d_parameters.get("EXC_Freq_High"))

        lowfreq = float(d_parameters.get("EXC_Freq_Low"))

        # CR for compatibility with Apex format as there is no EXciteSweep file
        if not highfreq and lowfreq:

            excitation_sweep_filelocation = self.locate_file(
                self.d_directory_location, "ExciteSweep"
            )
            lowfreq, highfreq = self.get_excite_sweep_range(
                excitation_sweep_filelocation
            )
            d_parameters["EXC_Freq_High"] = highfreq
            d_parameters["EXC_Freq_Low"] = lowfreq

    @staticmethod
    def get_excite_sweep_range(filename):
        """
        Function that returns the lower and higher frequency of the pulse generator
        """
        ExciteSweep_lines = genfromtxt(filename, comments="*", delimiter="\n")
        # CR ready if we need the full array
        highfreq = fromstring(ExciteSweep_lines[0])
        lowfreq = fromstring(ExciteSweep_lines[-1])

        return lowfreq[0], highfreq[0]

    @staticmethod
    def locate_file(folder, type_file_name):
        
        from pathlib import Path
        """
            type_of_file = ExciteSweep or apexAcquisition.method
            From the given folder this function return the absolute path to the ExciteSweep file, or the apexAcquisition.method file
            It should always be in a subfolder 
        
            ''' this code is a adaptation/automation of spike library
                https://bitbucket.org/delsuc/spike/
                from Marc Delsuc
            '''
        """
        
        directory_location = folder.glob( '**/*apexAcquisition.method')
        result = list(directory_location)
        if len(result) > 1:

            raise Exception(
                "You have more than 1 %s file in the %s folder, using the first one"
                % (type_file_name, folder)
            )

        elif len(result) == 0:

            raise Exception(
                "You don't have any %s file in the  %s folder, please double check the path"
                % (type_file_name, folder)
            )

        return result[0]

    @staticmethod
    def parse_parameters(parameters_filename):
        """ 
            TODO: change to beautiful soup xml parsing
            Open the given file and retrieve all parameters from apexAcquisition.method
            None is written when no value for value is found

            structure : <param name = "AMS_ActiveExclusion"><value>0</value></param>

            read_param returns  values in a dictionnary
            xml should be extinct, just a random option

            ''' this code is a adaptation/automation of spike library
                https://bitbucket.org/delsuc/spike/
                from Marc Delsuc
            '''
        """
        xmldoc = minidom.parse(parameters_filename.open())

        x = xmldoc.documentElement
        parameter_dict = {}
        children = x.childNodes
        for child in children:
            # print( child.node)
            if child.nodeName == "reportinfo":
                sections = child.childNodes
                for section in sections:
                    if section.nodeName == "section":
                        if section.getAttribute("title") == "Main":
                            for element in section.childNodes:
                                if element.nodeName == "section":
                                    if element.getAttribute("title") == "Polarity":
                                        if (
                                            str(
                                                element.childNodes[1].getAttribute(
                                                    "value"
                                                )
                                            )
                                            == "Negative"
                                        ):
                                            parameter_dict["Polarity"] = -1
                                        else:
                                            parameter_dict["Polarity"] = 1

            if child.nodeName == "paramlist":
                params = child.childNodes
                for param in params:
                    # print( param.nodeName)
                    if param.nodeName == "param":
                        paramenter_label = str(param.getAttribute("name"))
                        for element in param.childNodes:
                            if element.nodeName == "value":
                                try:
                                    parameter_value = str(element.firstChild.toxml())
                                    # print v
                                except:
                                    parameter_value = None

                            parameter_dict[paramenter_label] = parameter_value

        return parameter_dict





from numpy import array, polyfit, poly1d, where, trapz

from bisect import bisect_left

__author__ = "Yuri E. Corilo"
__date__ = "March 11, 2020"

class GCPeakCalculation(object):

    '''
    classdocs
    '''
    def calc_area(self, tic, dx):
        
        yy = tic[self.start_index:self.final_index]
        
        self._area = trapz(yy, dx = dx)

    def linear_ri(self, right_ri, left_ri,left_rt,right_rt ):

        #ri = ((right_ri -  left_ri) * (self.rt - left_rt)/ (right_rt - left_rt)) + left_ri   

        return left_ri + ((right_ri -  left_ri) * (self.rt - left_rt)/(right_rt-left_rt))

    def calc_ri(self, rt_ri_pairs):
        current_rt = self.rt

        rts = [rt_ri[0] for rt_ri in rt_ri_pairs]
        index = bisect_left(rts, current_rt)
        
        if index >= len(rt_ri_pairs):
            index -= 1
        
        current_ref = rt_ri_pairs[index]
        
        if current_rt == current_ref[0]:
            #print(current_rt, current_ref)
            self._ri = current_ref[1]
            return 1
        
        else:
            if index == 0:
                index += 1
            
            left_rt = rt_ri_pairs[index-1][0]
            left_ri = rt_ri_pairs[index-1][1]

            right_rt = rt_ri_pairs[index][0]
            right_ri =rt_ri_pairs[index][1]

            self._ri = self.linear_ri(right_ri, left_ri,left_rt,right_rt)

            #print(rt_ri_pairs[index-1], current_rt, rt_ri_pairs[index], left_ri, self._ri, right_ri)

            return 1


__author__ = "Yuri E. Corilo"
__date__ = "Oct 29, 2019"

from threading import Thread
from pathlib import Path
from io import BytesIO

import h5py
from s3path import S3Path

from corems.encapsulation.constant import Labels
from corems.mass_spectrum.factory.MassSpectrumClasses import MassSpecProfile
from corems.mass_spectra.factory.LC_Class import LCMSBase
from corems.encapsulation.factory.parameters import default_parameters

class ReadHDF_BoosterMassSpectra(Thread):
    
    '''class docs'''
    
    def __init__(self, file_location, analyzer='ICR', instrument_label='21T', auto_process=True):

        '''
		 # Parameters
		----------
		## file_location : Path or S3Path
        file_location is full data path
		'''
        Thread.__init__(self)
		
        if isinstance(file_location, str):
			# if obj is a string it defaults to create a Path obj, pass the S3Path if needed
            self.file_location = Path(file_location)

        self.lcms = LCMSBase(file_location, analyzer=analyzer,instrument_label=instrument_label)

        if isinstance(file_location, S3Path):
            data = BytesIO(file_location.open('rb').read())
        else:
            data = file_location
        
        self.hdf_obj =  h5py.File(data, 'r')

        self.list_scans =  sorted([int(i) for i in list(self.hdf_obj.keys())])

        self.initial_scan_number = self.list_scans [0]

        self.final_scan_number = self.list_scans [-1]

        self.file_location = file_location

        self.auto_process = True

        self.analyzer = analyzer

        self.instrument_label = instrument_label
    
    def get_polarity(self, file_location, scan):
        
        if isinstance(file_location, S3Path):
            data = BytesIO(file_location.open('rb').read())
        else:
            data = file_location
        
        self.h5pydata = h5py.File(data, 'r')

        self.scans = list(self.h5pydata.keys())
        
        polarity = self.get_attr_data(scan,'r_h_polarity')
        
        if polarity == 'negative scan': return -1
        
        else: return +1    
    
    def get_attr_data(self, scan, attr_srt):

        return self.hdf_obj[str(scan)].attrs[attr_srt]

    def import_mass_spectra(self, d_params):
        
        list_rt, list_tic = list(), list()
        
        for scan_number in self.list_scans:
            
            d_params["rt"] = list_rt.append(self.get_attr_data(scan_number, 'r_h_start_time'))

            d_params["scan_number"] = scan_number

            d_params['label'] = Labels.booster_profile
    
            d_params["polarity"] = self.get_polarity(self.file_location, scan_number)

            d_params["Aterm"] = self.get_attr_data(scan_number, 'r_cparams')[0]

            d_params["Bterm"] = self.get_attr_data(scan_number, 'r_cparams')[1]

            d_params['analyzer'] = self.analyzer
        
            d_params['instrument_label'] = self.instrument_label

            list_rt.append(d_params["rt"])

            list_tic.append(self.get_attr_data(scan_number, 'r_h_tic'))
            
            mass_spec = self.get_mass_spectrum(scan_number, d_params)

            self.lcms.add_mass_spectrum(mass_spec)

        self.lcms.retention_time = list_rt
        self.lcms.tic = list_tic
        self.lcms.scans_number = self.list_scans
        
    def get_mass_spectrum(self, scan, d_params):
        
        booster_data = self.hdf_obj[str(scan)]
        
        if booster_data.shape[0] is not 2:
            
            raise NotImplementedError('opening transient, needs read raw file here, get bandwidth, create transient class and then the mass spectrum')
       
        else:
            
            data_dict = {
                Labels.mz: booster_data[0],
                Labels.abundance: booster_data[1],
                Labels.rp: None,
                Labels.s2n: None,
            }
            
           
            mass_spec = MassSpecProfile(data_dict, d_params, auto_process=self.auto_process)

        return mass_spec

    def run(self):
        '''creates the lcms obj'''

        d_parameters = default_parameters(self.file_location)
        self.import_mass_spectra(d_parameters)
            
    def get_lcms_obj(self):
        
        if self.lcms.get(self.initial_scan_number):
            return self.lcms
        else:
            raise Exception("returning a empty lcms class")
    


__author__ = "Yuri E. Corilo"
__date__ = "Nov 06, 2019"


import csv
from pathlib import Path

from numpy import  NaN, concatenate
from openpyxl import load_workbook
from pandas import DataFrame, ExcelWriter, read_excel

from corems.mass_spectrum.output.export import HighResMassSpecExport
from corems.molecular_id.calc.SpectralSimilarity import methods_name
from corems.encapsulation.constant import Atoms
from corems.encapsulation.output import parameter_to_dict
from corems.mass_spectrum.factory.MassSpectrumClasses import MassSpecfromFreq
from corems import __version__, corems_md5
import uuid

class LowResGCMSExport():
    
    def __init__(self, out_file_path, gcms):
        '''
        output_type: str
            'excel', 'csv', 'hdf5' or 'pandas'
        '''
        
        self.output_file = Path(out_file_path)

        self.gcms = gcms

        self._init_columns()
        
    def _init_columns(self):

        columns =  ['Sample name', 'Peak Index',  'Retention Time', 'Retention Time Ref', 'Peak Height',
                'Peak Area', 'Retention index', 'Retention index Ref','Retention Index Score',
                'Similarity Score',
                'Spectral Similarity Score',
                'Compound Name']
        
        if self.gcms.molecular_search_settings.exploratory_mode:
                
                columns.extend(['Weighted Cosine Correlation',
                                'Cosine Correlation',
                                'Stein Scott Similarity',
                                'Stein Scott Similarity Nist',
                                'Pearson Correlation',
                                'Spearman Correlation',
                                'Kendall Tau Correlation', 
                                'Euclidean Distance',
                                'Manhattan Distance',
                                'Jaccard Distance',
                                'DWT Correlation',
                                'DFT Correlation' ])

                columns.extend(list(methods_name.values()))
                                                
        
        return columns        
    
    def get_pandas_df(self, id_label="corems:"):

        columns = self._init_columns()

        dict_data_list = self.get_list_dict_data(self.gcms)

        df = DataFrame(dict_data_list, columns=columns)

        df.name = self.gcms.sample_name

        return df

    def get_json(self, nan=False, id_label="corems:"):

        import json

        dict_data_list = self.get_list_dict_data(self.gcms)

        return json.dumps(dict_data_list, sort_keys=False, indent=4, separators=(',', ': '))

    def to_pandas(self, id_label="corems:"):

        columns = self._init_columns() 

        dict_data_list = self.get_list_dict_data(self.gcms)

        df = DataFrame(dict_data_list, columns=columns)

        df.to_pickle(self.output_file.with_suffix('.pkl'))

        self.write_settings(self.output_file, self.gcms, id_label="corems:")

    def to_excel(self, write_mode='w', id_label="corems:"):

        out_put_path = self.output_file.with_suffix('.xlsx')

        columns = self._init_columns()

        dict_data_list = self.get_list_dict_data(self.gcms)

        df = DataFrame(dict_data_list, columns=columns)

        if write_mode == 'a' and out_put_path.exists():

            writer = ExcelWriter(out_put_path, engine='openpyxl')
            # try to open an existing workbook
            writer.book = load_workbook(out_put_path)
            # copy existing sheets
            writer.sheets = dict((ws.title, ws) for ws in writer.book.worksheets)
            # read existing file
            reader = read_excel(out_put_path)
            # write out the new sheet
            df.to_excel(writer, index=False, header=False, startrow=len(reader) + 1)

            writer.close()
        else:

            df.to_excel(self.output_file.with_suffix('.xlsx'), index=False, engine='openpyxl')

        self.write_settings(self.output_file, self.gcms, id_label="corems:")

    def to_csv(self, separate_output=False, id_label="corems:"):

        if separate_output:
            # set write mode to write
            # this mode will overwrite the file without warning
            write_mode = 'w'
        else:
            # set write mode to append
            write_mode = 'a'
        
        columns = self._init_columns() 
        
        dict_data_list = self.get_list_dict_data(self.gcms)

        out_put_path = self.output_file.with_suffix('.csv')

        write_header = not out_put_path.exists()
        
        try:
            with open(out_put_path, write_mode, newline='') as csvfile:
                writer = csv.DictWriter(csvfile, fieldnames=columns)
                if write_header:
                    writer.writeheader()
                for data in dict_data_list:
                    writer.writerow(data)
            
            self.write_settings(self.output_file, self.gcms, id_label=id_label)
        
        except IOError as ioerror:
            print(ioerror)                 
    
    def to_hdf(self, id_label="corems:"):
        
        # save sample at a time
        def add_compound(gc_peak, compound_obj):

            compound_group = peak_group.create_group(compound_obj.name.replace('/', ''))
            compound_group.attrs["retention_time"] = compound_obj.rt
            compound_group.attrs["retention_index"] = compound_obj.ri
            compound_group.attrs["retention_index_score"] = compound_obj.ri_score
            compound_group.attrs["spectral_similarity_score"] = compound_obj.spectral_similarity_score
            compound_group.attrs["similarity_score"] = compound_obj.similarity_score
            
            compond_mz = compound_group.create_dataset('mz', data=np.array(compound_obj.mz), dtype="f8")  
            compond_abundance = compound_group.create_dataset('abundance', data=np.array(compound_obj.abundance), dtype="f8")

            if self.gcms.molecular_search_settings.exploratory_mode:
                
                compound_group.attrs['Spectral Similarities'] =  json.dumps(compound_obj.spectral_similarity_scores, sort_keys=False, indent=4, separators=(',', ': '))
                
        import h5py
        import json
        import numpy as np
        from datetime import datetime, timezone

        output_path = self.output_file.with_suffix('.hdf5')

        with h5py.File(output_path, 'w') as hdf_handle:

            timenow = str(datetime.now(timezone.utc).strftime("%d/%m/%Y %H:%M:%S %Z"))
            hdf_handle.attrs['time_stamp'] = timenow
            hdf_handle.attrs['data_structure'] = 'gcms'
            hdf_handle.attrs['analyzer'] = self.gcms.analyzer
            hdf_handle.attrs['instrument_label'] = self.gcms.instrument_label

            hdf_handle.attrs['sample_id'] = "self.gcms.id"
            hdf_handle.attrs['sample_name'] = self.gcms.sample_name
            hdf_handle.attrs['input_data'] = str(self.gcms.file_location)
            hdf_handle.attrs['output_data'] = str(output_path)
            hdf_handle.attrs['output_data_id'] = id_label + uuid.uuid4().hex
            hdf_handle.attrs['corems_version'] = __version__

            hdf_handle.attrs["Stats"] = json.dumps(self.get_data_stats(self.gcms), sort_keys=False, indent=4, separators=(',', ': '))
            hdf_handle.attrs["Calibration"] = json.dumps(self.get_calibration_stats(self.gcms, id_label), sort_keys=False, indent=4, separators=(',', ': '))
            hdf_handle.attrs["Blank"] = json.dumps(self.get_blank_stats(self.gcms), sort_keys=False, indent=4, separators=(',', ': '))

            corems_dict_setting = parameter_to_dict.get_dict_data_gcms(self.gcms)
            hdf_handle.attrs["CoreMSParameters"] = json.dumps(corems_dict_setting, sort_keys=False, indent=4, separators=(',', ': '))
            
            scans_dataset = hdf_handle.create_dataset('scans', data=np.array(self.gcms.scans_number), dtype="f8")                
            rt_dataset = hdf_handle.create_dataset('rt', data=np.array(self.gcms.retention_time), dtype="f8")                
            tic_dataset = hdf_handle.create_dataset('tic', data=np.array(self.gcms.tic), dtype="f8")            
            processed_tic_dataset = hdf_handle.create_dataset('processed_tic', data=np.array(self.gcms.processed_tic), dtype="f8")

            output_score_method = self.gcms.molecular_search_settings.output_score_method

            for gc_peak in self.gcms:

                # print(gc_peak.rt)
                # print(gc_peak.tic)
                
                # check if there is a compound candidate 
                peak_group = hdf_handle.create_group(str(gc_peak.rt))
                peak_group.attrs["deconvolution"] = int(self.gcms.chromatogram_settings.use_deconvolution)
                
                peak_group.attrs["start_index"] = gc_peak.start_index
                peak_group.attrs["index"] = gc_peak.index
                peak_group.attrs["final_index"] = gc_peak.final_index

                peak_group.attrs["retention_index"] = gc_peak.ri
                peak_group.attrs["retention_time"] = gc_peak.rt
                peak_group.attrs["area"] = gc_peak.area
                
                mz = peak_group.create_dataset('mz', data=np.array(gc_peak.mass_spectrum.mz_exp), dtype="f8")
                abundance = peak_group.create_dataset('abundance', data=np.array(gc_peak.mass_spectrum.abundance), dtype="f8")

                if gc_peak:

                    if output_score_method == 'highest_sim_score':
                        compound_obj = gc_peak.highest_score_compound
                        add_compound(gc_peak, compound_obj)

                    elif output_score_method == 'highest_ss':
                        compound_obj = gc_peak.highest_ss_compound
                        add_compound(gc_peak, compound_obj)

                    else:

                        for compound_obj in gc_peak:
                            add_compound(gc_peak, compound_obj)
                    
    def get_data_stats(self, gcms):
        
        matched_peaks = gcms.matched_peaks
        no_matched_peaks = gcms.no_matched_peaks
        unique_metabolites = gcms.unique_metabolites
        
        peak_matchs_above_0p85 = 0
        unique_peak_match_above_0p85 = 0
        for match_peak in matched_peaks:
            gc_peak_above_85 = 0
            matches_above_85 = list(filter(lambda m: m.similarity_score >= 0.85, match_peak))
            if matches_above_85:
                peak_matchs_above_0p85 +=1
            if len(matches_above_85) == 1:
                unique_peak_match_above_0p85 += 1

        data_stats = {}
        data_stats['average_signal_noise'] = "ni"
        data_stats['chromatogram_dynamic_range'] = gcms.dynamic_range
        data_stats['total_number_peaks'] = len(gcms)
        data_stats['total_peaks_matched'] = len(matched_peaks)
        data_stats['total_peaks_without_matches'] = len(no_matched_peaks)
        data_stats['total_matches_above_similarity_score_0.85'] = peak_matchs_above_0p85
        data_stats['single_matches_above_similarity_score_0.85'] = unique_peak_match_above_0p85
        data_stats['unique_metabolites'] = len(unique_metabolites)

        return data_stats

    def get_calibration_stats(self, gcms, id_label):
        
        calibration_parameters = {}

        calibration_parameters['calibration_rt_ri_pairs_ref'] = gcms.ri_pairs_ref
        calibration_parameters['file_path'] = str(gcms.cal_file_path)
        calibration_parameters['file_id'] = id_label + corems_md5(gcms.cal_file_path)
        calibration_parameters['sample_name'] = str(gcms.cal_file_path.stem)
        calibration_parameters['calibration_method'] = ""

        return calibration_parameters

    def get_blank_stats(self, gcms):
        
        blank_parameters = {}

        blank_parameters['sample_name'] = "ni"
        blank_parameters['blank_id'] = "ni"
        blank_parameters['file_path'] = "ni"
        blank_parameters['file_id'] = "ni"
        blank_parameters['common_features_to_blank'] = "ni"

        return blank_parameters

    def get_parameters_json(self, gcms, id_label, output_path):

        import json

        if isinstance(output_path, str):
            output_path = Path(output_path)

        paramaters_path = output_path.with_suffix('.json')

        if paramaters_path.exists():
            with paramaters_path.open() as current_param:
                output_parameters_dict = json.load(current_param)

        else:
            output_parameters_dict = {}
      
            output_parameters_dict['sample_name'] = []
            output_parameters_dict['sample_id'] = []
            output_parameters_dict['input_data'] = []
            output_parameters_dict['input_data_id'] = []

        blank_parameters = {}
        
        output_parameters_dict['analyzer'] = gcms.analyzer
        output_parameters_dict['instrument_label'] = gcms.instrument_label
        
        output_parameters_dict['sample_name'].append(gcms.sample_name)
        output_parameters_dict['sample_id'].append("sample_id")
        output_parameters_dict['input_data'].append(str(gcms.file_location))
        output_parameters_dict['input_data_id'].append(id_label + corems_md5(gcms.file_location))
        
        output_parameters_dict['output_data'] = str(output_path)
        output_parameters_dict['output_data_id'] = id_label + uuid.uuid4().hex
        output_parameters_dict['corems_version'] = __version__
        
        output_parameters_dict["Stats"] = self.get_data_stats(gcms)
        output_parameters_dict["Calibration"] = self.get_calibration_stats(gcms, id_label)
        output_parameters_dict["Blank"] = self.get_blank_stats(gcms)

        corems_dict_setting = parameter_to_dict.get_dict_data_gcms(gcms)
        output_parameters_dict["CoreMSParameters"] = corems_dict_setting

        output = json.dumps(output_parameters_dict, sort_keys=False, indent=4, separators=(',', ': '))

        return output

    def write_settings(self, output_path, gcms, id_label="corems:"):

        output = self.get_parameters_json(gcms, id_label, output_path)

        with open(output_path.with_suffix('.json'), 'w', encoding='utf8', ) as outfile:

            outfile.write(output)

    def get_list_dict_data(self, gcms, include_no_match=True, no_match_inline=False) :

        output_score_method = gcms.molecular_search_settings.output_score_method

        dict_data_list = []

        def add_match_dict_data():

            out_dict = {'Sample name': gcms.sample_name,
                        'Peak Index': gcpeak_index, 
                        'Retention Time': gc_peak.rt,
                        'Retention Time Ref': compound_obj.rt,
                        'Peak Height': gc_peak.tic,
                        'Peak Area': gc_peak.area,
                        'Retention index': gc_peak.ri,
                        'Retention index Ref': compound_obj.ri,
                        'Retention Index Score': compound_obj.ri_score,
                        'Spectral Similarity Score': compound_obj.spectral_similarity_score,
                        'Similarity Score': compound_obj.similarity_score,
                        'Compound Name': compound_obj.name
                        }

            if self.gcms.molecular_search_settings.exploratory_mode:

                out_dict.update({
                    'Weighted Cosine Correlation': compound_obj.spectral_similarity_scores.get("weighted_cosine_correlation"), 
                    'Cosine Correlation': compound_obj.spectral_similarity_scores.get("cosine_correlation"), 
                    'Stein Scott Similarity': compound_obj.spectral_similarity_scores.get("stein_scott_similarity"), 
                    'Stein Scott Similarity Nist': compound_obj.spectral_similarity_scores.get("stein_scott_similarity_nist"),                     
                    'Pearson Correlation': compound_obj.spectral_similarity_scores.get("pearson_correlation"), 
                    'Spearman Correlation': compound_obj.spectral_similarity_scores.get("spearman_correlation"), 
                    'Kendall Tau Correlation': compound_obj.spectral_similarity_scores.get("kendall_tau_correlation"), 
                    'DFT Correlation': compound_obj.spectral_similarity_scores.get("dft_correlation"),
                    'DWT Correlation': compound_obj.spectral_similarity_scores.get("dwt_correlation"),
                    'Euclidean Distance': compound_obj.spectral_similarity_scores.get("euclidean_distance"), 
                    'Manhattan Distance': compound_obj.spectral_similarity_scores.get("manhattan_distance"),
                    'Jaccard Distance': compound_obj.spectral_similarity_scores.get("jaccard_distance"),
                })
                
                for method in methods_name:
                    
                    out_dict[methods_name.get(method)] = compound_obj.spectral_similarity_scores.get(method)

            dict_data_list.append(out_dict)
        
        def add_no_match_dict_data():

            dict_data_list.append({'Sample name': gcms.sample_name,
                                    'Peak Index': gcpeak_index,
                                    'Retention Time': gc_peak.rt,
                                    'Peak Height': gc_peak.tic,
                                    'Peak Area': gc_peak.area,
                                    'Retention index': gc_peak.ri,
                                  })

        for gcpeak_index, gc_peak in enumerate(gcms.sorted_gcpeaks):

            # check if there is a compound candidate 
            if gc_peak:
                
                if output_score_method == 'highest_sim_score':
                    compound_obj = gc_peak.highest_score_compound
                    add_match_dict_data()

                elif output_score_method == 'highest_ss':
                    
                    compound_obj = gc_peak.highest_ss_compound
                    add_match_dict_data()
                    
                else:
                    for compound_obj in gc_peak:
                        add_match_dict_data()  # add monoisotopic peak

            else:
                # include not_match
                if include_no_match and no_match_inline:
                    add_no_match_dict_data()

        if include_no_match and not no_match_inline:
            for gcpeak_index , gc_peak in enumerate(gcms.sorted_gcpeaks):
                if not gc_peak:
                    add_no_match_dict_data()
        
        return dict_data_list

    
class HighResMassSpectraExport(HighResMassSpecExport):
    '''
    
    '''
    def __init__(self, out_file_path, mass_spectra, output_type='excel'):
        '''
        output_type: str
            'excel', 'csv', 'hdf5' or 'pandas'
        '''
        
        self.output_file = Path(out_file_path)

        self.dir_loc = Path(out_file_path + ".corems")
        
        self.dir_loc.mkdir(exist_ok=True)

        # 'excel', 'csv' or 'pandas'
        self._output_type = output_type

        self.mass_spectra = mass_spectra

        self._init_columns()
        
    def get_pandas_df(self):

        list_df = []
        
        for mass_spectrum in self.mass_spectra:

            columns = self.columns_label + self.get_all_used_atoms_in_order(mass_spectrum)
            
            dict_data_list = self.get_list_dict_data(mass_spectrum)
            
            df = DataFrame(dict_data_list, columns=columns)
            
            scan_number = mass_spectrum.scan_number
            
            df.name = str(self.output_file) + '_' + str(scan_number)
            
            list_df.append(df)
        
        return list_df

    def to_pandas(self):
        
        for mass_spectrum in self.mass_spectra:

            columns = self.columns_label + self.get_all_used_atoms_in_order(mass_spectrum)

            dict_data_list = self.get_list_dict_data(mass_spectrum)

            df = DataFrame(dict_data_list, columns=columns)

            scan_number = mass_spectrum.scan_number

            out_filename = Path("%s_scan%s%s" % (self.output_file, str(scan_number), '.pkl'))
            
            df.to_pickle(self.dir_loc / out_filename)

            self.write_settings(self.dir_loc / out_filename.with_suffix(''), mass_spectrum)

    def to_excel(self):

        for mass_spectrum in self.mass_spectra:
            
            columns = self.columns_label + self.get_all_used_atoms_in_order(mass_spectrum)

            dict_data_list = self.get_list_dict_data(mass_spectrum)

            df = DataFrame(dict_data_list, columns=columns)

            scan_number = mass_spectrum.scan_number
            
            out_filename = Path("%s_scan%s%s" % (self.output_file, str(scan_number), '.xlsx'))
            
            df.to_excel(self.dir_loc / out_filename)

            self.write_settings(self.dir_loc / out_filename.with_suffix(''), mass_spectrum)
            

    def to_csv(self):
        
        import csv

        for mass_spectrum in self.mass_spectra:
            
            columns = self.columns_label + self.get_all_used_atoms_in_order(mass_spectrum)

            scan_number = mass_spectrum.scan_number

            dict_data_list = self.get_list_dict_data(mass_spectrum)

            out_filename = Path("%s_scan%s%s" % (self.output_file, str(scan_number), '.csv'))
            
            
            with open(self.dir_loc / out_filename, 'w', newline='') as csvfile:
                writer = csv.DictWriter(csvfile, fieldnames=columns)
                writer.writeheader()
                for data in dict_data_list:
                    writer.writerow(data)

            self.write_settings(self.dir_loc / out_filename.with_suffix(''), mass_spectrum)                    
    
    def get_mass_spectra_attrs(self, mass_spectra):

        dict_ms_attrs = {}
        dict_ms_attrs['analyzer'] = self.mass_spectra.analyzer
        dict_ms_attrs['instrument_label'] = self.mass_spectra.instrument_label
        dict_ms_attrs['sample_name'] = self.mass_spectra.sample_name
        
        import json  
        return json.dumps(dict_ms_attrs, sort_keys=False, indent=4, separators=(',', ': '))

    def to_hdf(self):
        
        import h5py
        import json
        from numpy import array
        from datetime import datetime, timezone

        with h5py.File(self.output_file.with_suffix('.hdf5'), 'a') as hdf_handle:
            
            if not hdf_handle.attrs.get('date_utc'):

                    timenow = str(datetime.now(timezone.utc).strftime("%d/%m/%Y %H:%M:%S %Z"))
                    hdf_handle.attrs['date_utc'] = timenow
                    hdf_handle.attrs['filename'] = self.mass_spectra.file_location.name
                    hdf_handle.attrs['data_structure'] = 'mass_spectra'
                    hdf_handle.attrs['analyzer'] = self.mass_spectra.analyzer
                    hdf_handle.attrs['instrument_label'] = self.mass_spectra.instrument_label
                    hdf_handle.attrs['sample_name'] = self.mass_spectra.sample_name

            for mass_spectrum in self.mass_spectra:

                list_results = self.list_dict_to_list(mass_spectrum, is_hdf5=True)
                
                dict_ms_attrs = self.get_mass_spec_attrs(mass_spectrum)
                
                setting_dicts = parameter_to_dict.get_dict_data_ms(mass_spectrum)

                columns_labels = json.dumps(self.columns_label + self.get_all_used_atoms_in_order(mass_spectrum))

                group_key = str(mass_spectrum.scan_number)

                if not group_key in hdf_handle.keys():
                
                    scan_group = hdf_handle.create_group(str(mass_spectrum.scan_number))

                    if list(mass_spectrum.abundance_profile):
                    
                       mz_abun_array = concatenate((mass_spectrum.abundance_profile, mass_spectrum.mz_exp_profile), axis=0)
                        
                       
                       raw_ms_dataset = scan_group.create_dataset('raw_ms', data=mz_abun_array, dtype="f8")

                    else:
                        #create empy dataset for missing raw data
                        raw_ms_dataset = scan_group.create_dataset('raw_ms', dtype="f8")

                    raw_ms_dataset.attrs['MassSpecAttrs'] = json.dumps(dict_ms_attrs, sort_keys=False, indent=4, separators=(',', ': '))
                    
                    if isinstance(mass_spectrum, MassSpecfromFreq):
                        raw_ms_dataset.attrs['TransientSetting'] = json.dumps(setting_dicts.get('TransientSetting'), sort_keys=False, indent=4, separators=(',', ': '))

                else:
                    
                    scan_group = hdf_handle.get(group_key)
        
                    # if there is not processed data len = 0, otherwise len() will return next index
                index_processed_data = str(len(scan_group.keys()))
                
                timenow = str(datetime.now(timezone.utc).strftime("%d/%m/%Y %H:%M:%S %Z"))
                
                processed_dset = scan_group.create_dataset(index_processed_data, data=list_results)
            
                processed_dset.attrs['date_utc'] = timenow

                processed_dset.attrs['ColumnsLabels'] = columns_labels
                
                processed_dset.attrs['MoleculaSearchSetting'] = json.dumps(setting_dicts.get('MoleculaSearch'), sort_keys=False, indent=4, separators=(',', ': '))
                
                processed_dset.attrs['MassSpecPeakSetting'] = json.dumps(setting_dicts.get('MassSpecPeak'), sort_keys=False, indent=4, separators=(',', ': '))

                processed_dset.attrs['MassSpectrumSetting'] = json.dumps(setting_dicts.get('MassSpectrum'), sort_keys=False, indent=4, separators=(',', ': '))


from corems.molecular_id.search.compoundSearch import LowResMassSpectralMatch


def get_rt_ri_pairs(gcms_ref_obj, sql_obj=None):
    
    lowResSearch = LowResMassSpectralMatch(gcms_ref_obj,sql_obj=sql_obj, calibration=True)

    lowResSearch.run()

    dict_ri_rt = {}

    list_of_compound_obj = {}

    for gcms_peak in gcms_ref_obj:

        # has a compound matched
        if gcms_peak:
            
            compound_obj = gcms_peak.highest_ss_compound
            
            if not compound_obj.ri in dict_ri_rt.keys():
                
                dict_ri_rt[compound_obj.ri] = [(gcms_peak.mass_spectrum.rt, compound_obj)]

            else:
                
                dict_ri_rt[compound_obj.ri].append((gcms_peak.mass_spectrum.rt, compound_obj))
            
            print(compound_obj.name, gcms_peak.mass_spectrum.rt, compound_obj.spectral_similarity_score)
    
    ris = [i for i in  dict_ri_rt.keys()]
    rts = [max(i, key = lambda c: c[1].spectral_similarity_score)[0] for i in dict_ri_rt.values()]
    
    rt_ri_pairs = list(zip(rts, ris)) 
    
    return rt_ri_pairs


from corems.encapsulation.factory.parameters import MSParameters, GCMSParameters

def get_dict_all_default_data():
    
    return { "MolecularFormulaSearch": MSParameters.molecular_search.__dict__,
             "Transient": MSParameters.transient.__dict__,
             "MassSpectrum": MSParameters.mass_spectrum.__dict__,
             "MassSpecPeak": MSParameters.ms_peak.__dict__,
             "DataInput": MSParameters.data_input.__dict__,
             "MolecularSearch": GCMSParameters.molecular_search.__dict__,
             "GasChromatograph": GCMSParameters.gc_ms.__dict__,
            }

def get_dict_data_ms(mass_spec):

    if mass_spec._transient_settings:

        return { "MolecularFormulaSearch": mass_spec.molecular_search_settings.__dict__,
                "Transient": mass_spec.transient_settings.__dict__,
                "MassSpectrum": mass_spec.settings.__dict__,
                "MassSpecPeak": mass_spec.mspeaks_settings.__dict__
                }
    else:
        
        return { "MolecularFormulaSearch": mass_spec.molecular_search_settings.__dict__,
                "MassSpectrum": mass_spec.settings.__dict__,
                "MassSpecPeak": mass_spec.mspeaks_settings.__dict__
                }
                
def get_dict_ms_default_data():
    
    return { "MolecularFormulaSearch": MSParameters.molecular_search.__dict__,
             "Transient": MSParameters.transient.__dict__,
             "MassSpectrum": MSParameters.mass_spectrum.__dict__,
             "MassSpecPeak": MSParameters.ms_peak.__dict__,
             "DataInput": MSParameters.data_input.__dict__,
            }

def get_dict_gcms_default_data():
    
    return { "MolecularSearch": GCMSParameters.molecular_search.__dict__,
             "GasChromatograph": GCMSParameters.gc_ms.__dict__,
            }

def get_dict_data_gcms(gcms):

    return { "MolecularSearch": gcms.molecular_search_settings.__dict__,
             "GasChromatograph":  gcms.chromatogram_settings.__dict__,
            }          


from os import path

from numpy import dtype, fromfile


__author__ = "Yuri E. Corilo"
__date__ = "Jun 19, 2019"



class ReadMidasDatFile():
    '''
    Reads Midas data files, works for both Predator Analysis data and Thermo DataStation
    '''

    def __init__(self, filename_path):
        '''
        Constructor
        '''
        raise NotImplementedError("This class is not yet implemented, if you want to use it please contact the author at corilo@pnnl.gov or feel free to implement it")
        if not path.isfile(filename_path):
            raise Exception("File does not exist: "+ filename_path)
        
        self.filename_path = filename_path
        
    def read_file(self):

        data_file = open(self.filename_path, 'rb')

        # modo_de_ions = "POSITIVE ION MODE"
        d_params = self.parse_parameters(self.parameter_filename_location)
        
        transient_data = self.get_transient_data(data_file, d_params, d_params)
        
        return transient_data, d_params
        
    
    def get_transient_data(self, data_file, d_params):

        #dt = np.dtype('<f')
        if d_params.get("storage_type").split()[0] == "int":
            dt = dtype('i2')
            
        else:
            dt = dtype('<f')    
        #dt = np.dtype(int)
        
        myarray = fromfile(data_file, dtype=dt)
        
        data_file.close()

        if d_params.get("storage_type").split()[0] == "int":
            return myarray * d_params.get("VoltageScale")
        
        else:
            
            return myarray     
        
    def parse_parameter(self, f):
        
        output_parameters = {}
        output_parameters["filename_path"] = self.d_directory_location
        
        line = f.readline()
        
        while line != "Data:\n":

            if line[0:8] == "highfreq":
                final_frequency = float(line.split(":")[1])
                output_parameters["exc_high_freq"] = final_frequency
                
            elif line[0:7] == "lowfreq":
                initial_frequency = float(line.split(":")[1])
                output_parameters["exc_low_freq"] = initial_frequency
                
            elif line[0:9] == "sweeprate":
                sweeprate = float(line.split(":")[1])
               
                output_parameters['sweeprate'] = sweeprate
                
            elif line[0:13] == "Source Coeff0":
                Acoef = float(line.split(":")[1])
                output_parameters["Aterm"] = Acoef
                #print f.readline()
            elif line[0:13] == "Source Coeff1":
                output_parameters["Bterm"] = "Bcoef"
                
            elif line[0:13] == "Voltage Scale":
                voltage_scale = float(line.split(":")[1])
                output_parameters["VoltageScale"] = voltage_scale
                
            elif line[0:9] == "Bandwidth":
                bandwidth = float(line.split(":")[1])
                output_parameters["bandwidth"] = bandwidth
                
            elif line[0:11] == "Data Points":
                datapoints = float(line.split(":")[1])
                output_parameters["number_data_points"] = datapoints
                
            elif line[0:12] == "Storage Type":
                storage_type = line.split(":")[1]
                output_parameters["storage_type"] = storage_type
                
            elif line[0:12] == "Trap Voltage":
                trap_voltage = float(line.split(":")[1])
                #Bcoef = Bcoef*trap_voltage
                output_parameters["trap_voltage"] = trap_voltage
                
            line = f.readline()
            
        return output_parameters




__author__ = "Yuri E. Corilo"
__date__ = "Feb 12, 2020"

import os 
from dataclasses import dataclass

from sqlalchemy import create_engine, Column, Integer, String, Float,  exists, Binary
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound
from sqlalchemy.pool import QueuePool
from sqlalchemy import between

from numpy import array, frombuffer

Base = declarative_base()

@dataclass
class LowResCompoundRef:
    "this class is use to store the results inside the GCPeak class"
    def __init__(self, compounds_dict):
        
        self.name = compounds_dict.get("id")
        self.ri = compounds_dict.get("ri")
        self.rt = compounds_dict.get("rt")
        self.casno  = compounds_dict.get("casno")
        self.comment  = compounds_dict.get("comment")
        self.peaks_count = compounds_dict.get("peaks_count")
        
        self.mz  = compounds_dict.get('mz') 
        self.abundance  = compounds_dict.get("abundance") 

        self.source_temp_c  = compounds_dict.get("source_temp_c") 
        self.ev  = compounds_dict.get("ev") 
        self.formula  = compounds_dict.get("formula") 
        self.source = compounds_dict.get("source") 

        self.similarity_score = None
        self.ri_score = None
        self.spectral_similarity_score = None
        self.spectral_similarity_scores = {}

class LowResolutionEICompound(Base):  
    
    __tablename__ = 'pnnl_lowres_gcms_compounds'

    id = Column( String, primary_key=True)
    
    name = Column(String, nullable=False)
    formula = Column(String, nullable=True)
    ri = Column(Float, nullable=False)
    rt = Column(Float, nullable=False)
    
    source = Column(String, nullable=True)
    casno = Column(String, nullable=False)
    comment = Column(String, nullable=True)
    
    source_temp_c =  Column(Float, nullable=True)
    ev = Column(Float, nullable=True)

    peaks_count = Column(Integer, nullable=False)
    mz = Column(Binary,  nullable=False)
    abundance = Column(Binary,  nullable=False)

    def __init__(self, dict_data): 
        
        self.id = dict_data.get('NAME')
        
        self.name =dict_data.get('NAME')
        self.formula = dict_data.get('FORM')
        self.ri =dict_data.get('RI')
        self.rt = dict_data.get('RT')

        self.source = dict_data.get('SOURCE')
        self.casno =dict_data.get('CASNO')
        self.comment = dict_data.get('COMMENT')

        self.peaks_count = dict_data.get('NUM PEAKS')
        
        # mz and abun are numpy arrays of 64 bits integer
        # when using postgres array might be a better option
        
        self.mz = array(dict_data.get('mz'), dtype='int32').tobytes()
        self.abundance = array(dict_data.get('abundance'), dtype="int32").tobytes()
        
    def __repr__(self):
        return "<LowResolutionEICompound(name= %s , cas number = %s, formula = %s, Retention index= %.1f, Retention time= %.1f comment='%i')>" % (
                                    self.name, self.casno, self.formula, self.ri, self.rt, self.comment)

class EI_LowRes_SQLite:
    
    def __init__(self, url='sqlite://'):

        self.engine = self.init_engine(url)

        Base.metadata.create_all(self.engine)

        Session = sessionmaker(bind=self.engine)

        self.session = Session()

    def __exit__(self, exc_type, exc_val, exc_tb):
        # make sure the db connection gets closed
        # 
        self.commit()
        self.session.close()
        self.engine.dispose()

    def init_engine(self, url):
        
        directory = os.getcwd()
        
        if not url:
            
            if not os.path.isdir(directory + '/db'):
                    
                os.mkdir(directory + '/db')    
            
            url = 'sqlite:///{DB}/db/pnnl_lowres_gcms_compounds.sqlite'.format(DB=directory)

        return create_engine(url, poolclass=QueuePool)

    def __enter__(self):
        
        return self
    
    def add_compound_list(self, data_dict_list):
        
        #print(data_dict_list[0])
        for data_dict in data_dict_list:
            #print(data_dict.get('NUM PEAKS'))
            if not data_dict.get('NUM PEAKS'):
                data_dict['NUM PEAKS'] = len(data_dict.get('mz'))
            if not data_dict.get('CASNO'):
                data_dict['CASNO'] = data_dict.get('CAS')
                
        self.session.add_all( [LowResolutionEICompound(data_dict) for data_dict in data_dict_list] )
    
    def add_compound(self, data_dict): 

        one_compound = LowResolutionEICompound(data_dict)  
        self.session.add(one_compound)  
        self.commit()

    def commit(self):
        
        try:
            self.session.commit()  
        except SQLAlchemyError as e:
            self.session.rollback()
            print(str(e))
   
    def row_to_dict(self, row):
            
        row = {c.name: getattr(row, c.name) for c in row.__table__.columns}        
        
        row['mz'] = frombuffer(row.get('mz'), dtype="int32")
        row['abundance'] = frombuffer(row.get('abundance'), dtype="int32")

        return row

    def get_all(self,):

        compounds = self.session.query(LowResolutionEICompound).all()
       
        return [self.row_to_dict(compound) for compound in compounds]

    def query_min_max_rt(self, min_max_rt, ):
        
        min_rt, max_rt = min_max_rt

        compounds = self.session.query(LowResolutionEICompound).filter(LowResolutionEICompound.rt.between(min_rt, max_rt))    
        
        return [self.row_to_dict(compound) for compound in compounds]

    def query_min_max_ri(self, min_max_ri):
        
        min_ri, max_ri = min_max_ri

        compounds = self.session.query(LowResolutionEICompound).filter(LowResolutionEICompound.ri.between(min_ri, max_ri))    
        
        return [self.row_to_dict(compound) for compound in compounds]

    def query_names_and_rt(self, min_max_rt, compound_names):

        min_rt, max_rt = min_max_rt

        compounds = self.session.query(LowResolutionEICompound).filter(LowResolutionEICompound.name.in_(compound_names)).filter(
                                       LowResolutionEICompound.rt >= min_rt,
                                       LowResolutionEICompound.rt <= max_rt,
                                       )

        # self.session.query.select(LowResolutionEICompound).where(between(LowResolutionEICompound.ri, min_ri, max_ri))    
        x = [self.row_to_dict(compound) for compound in compounds]
        return [self.row_to_dict(compound) for compound in compounds]

    def query_min_max_ri_and_rt(self, min_max_ri, min_max_rt, ):
        
        min_ri, max_ri = min_max_ri
        
        min_rt, max_rt = min_max_rt

        compounds = self.session.query(LowResolutionEICompound).filter(
            LowResolutionEICompound.ri <= max_ri,
            LowResolutionEICompound.ri >= min_ri,
            LowResolutionEICompound.ri >= min_rt,
            LowResolutionEICompound.ri >= max_rt,
            )
        
        #self.session.query.select(LowResolutionEICompound).where(between(LowResolutionEICompound.ri, min_ri, max_ri))    
        
        return [self.row_to_dict(compound) for compound in compounds]

    def delete_compound(self, compound):
        
        try:
            self.session.delete(compound)  
            self.session.commit()  
        
        except SQLAlchemyError as e:
            self.session.rollback()
            print(str(e))

    def purge(self):
        '''Carefull, this will delete the entire database table'''
        self.session.query(LowResolutionEICompound).delete()
        self.session.commit()  

    def clear_data(self):
        '''clear tables'''
        meta = Base.metadata
        for table in reversed(meta.sorted_tables):
            print ('Clear table %s' % table)
            self.session.execute(table.delete())
        self.session.commit()
   